from abc import ABC, abstractmethod
from typing import List, Dict, Type, Optional, Union, Tuple
import gc

class DampingBase:
    _dampings = {}
    def __init__(self):
        tag = self._get_next_tag()
        self.name = f"damping{tag}"
        self.tag = tag
        self._dampings[tag] = self

    @classmethod
    def _get_next_tag(cls):
        """Returns the next available tag for a new damping"""
        return len(cls._dampings) + 1
    
    @classmethod
    def remove_damping(cls, tag):
        """Removes a damping from the list of dampings"""
        if tag in cls._dampings:
            del cls._dampings[tag]
            cls._update_tags()


    @staticmethod
    def get_damping(tag):
        """Returns the damping with the given tag"""
        return DampingBase._dampings[tag]
    

    def remove(self):
        """Removes the damping from the list of dampings"""
        refrences = gc.get_referrers(self)
        tag = self.tag
        self.__class__.remove_damping(tag)
        
    

    @classmethod
    def _update_tags(cls):
        """Updates the tags of all dampings"""
        tags = cls._dampings.keys()
        tags = sorted(tags)
        new_dampings = {}
        for i, tag in enumerate(tags):
            cls._dampings[tag].tag = i + 1
            cls._dampings[tag].name = f"damping{i + 1}"
            new_dampings[i + 1] = cls._dampings[tag]
        cls._dampings = new_dampings
        
    def __str__(self) -> str:
        res = f"Damping Class:\t{self.__class__.__name__}"
        res += f"\n\tName: {self.name}"
        res += f"\n\tTag: {self.tag}"
        return res
    
    @classmethod
    def print_dampings(cls):
        """Prints all the dampings"""
        print("Printing all dampings:")
        for tag in cls._dampings:
            print(cls._dampings[tag])

   

        


    @staticmethod
    def get_Parameters() -> List[Tuple[str, str]]:
        """Returns the parameters of the damping"""
        pass

    @staticmethod
    def get_Notes() -> Dict[str, Union[str, list]]:
        """Returns the notes of the damping"""
        pass
    
    @abstractmethod
    def get_values(self)-> Dict[str, Union[str, int, float, list]]:
        """Returns the values of the parameters of the damping"""
        pass

    @abstractmethod
    def update_values(self, **kwargs) -> None:
        """Updates the values of the parameters of the damping"""
        pass

    @abstractmethod
    def to_tcl(self) -> str:
        """Returns the TCL code of the damping"""
        pass
        





class RayleighDamping(DampingBase):
    def __init__(self, **kwargs):
        kwargs = self.validate(**kwargs)
        super().__init__()
        self.alphaM = kwargs["alphaM"]
        self.betaK = kwargs["betaK"]
        self.betaKInit = kwargs["betaKInit"]
        self.betaKComm = kwargs["betaKComm"]

    
    def __str__(self) -> str:
        res = super().__str__()
        res += f"\n\talphaM: {self.alphaM}"
        res += f"\n\tbetaK: {self.betaK}"
        res += f"\n\tbetaKInit: {self.betaKInit}"
        res += f"\n\tbetaKComm: {self.betaKComm}"
        return res



    @staticmethod
    def get_Parameters() -> List[Tuple[str, str]]:
        """Returns the parameters of the damping"""
        return [
            ("alphaM", "factor applied to elements or nodes mass matrix (optional, default=0.0)"),
            ("betaK", "factor applied to elements or nodes stiffness matrix (optional, default=0.0)"),
            ("betaKInit", "factor applied to elements initial stiffness matrix (optional, default=0.0)"),
            ("betaKComm", "factor applied to elements commited stiffness matrix (optional, default=0.0)")
        ]
    
    @staticmethod
    def get_Notes() -> Dict[str, Union[str, list]]:
        """Returns the notes of the damping"""
        return {
        "Notes": [
            "The damping matrix is calculated as: C = α<sub>M</sub> M + β<sub>K</sub>  K + β<sub>KInit</sub> Kinit + β<sub>KComm</sub>  Kcomm",
            "The usage of Rayleigh damping may provide incorrect result when used with Non-Linear Time History Analysis using Concentrated Plasticity Model."
            ],
        "References": [
            "https://opensees.github.io/OpenSeesDocumentation/user/manual/model/damping/rayleigh.html",
            "https://opensees.berkeley.edu/wiki/index.php/Rayleigh_Damping_Command"
        ]
        }
    
    @staticmethod
    def validate(**kwargs)-> Dict[str, Union[str, list]]:
        """Validates the damping"""
        newkwargs = {
            "alphaM": 0.0,
            "betaK": 0.0,
            "betaKInit": 0.0,
            "betaKComm": 0.0
        }


        #  check if the values are floats and are between 0 and 1
        for key in newkwargs:
            try:
                newkwargs[key] = float(kwargs.get(key, 0.0))
            except ValueError:
                raise ValueError(f"{key} should be a float")


        for key in newkwargs:
            try:
                if newkwargs[key] < 0 or newkwargs[key] > 1:
                    raise ValueError
            except ValueError:
                raise ValueError(f"{key} should be a float between 0 and 1")
        
        eps = 1e-10
        res = newkwargs["alphaM"] + newkwargs["betaK"] 
        res += newkwargs["betaKInit"] + newkwargs["betaKComm"]
        # at least one of the values should be greater than 0
        if res < eps:
            raise ValueError("At least one of the damping factors should be greater than 0")
        
        return newkwargs

    def get_values(self)-> Dict[str, Union[str, int, float, list]]:
        return {
            "alphaM": self.alphaM,
            "betaK": self.betaK,
            "betaKInit": self.betaKInit,
            "betaKComm": self.betaKComm
        }
    
    def update_values(self, **kwargs) -> None:
        """Updates the values of the parameters of the damping"""
        kwargs = self.validate(**kwargs)
        self.alphaM = kwargs["alphaM"]
        self.betaK = kwargs["betaK"]
        self.betaKInit = kwargs["betaKInit"]
        self.betaKComm = kwargs["betaKComm"]

    def to_tcl(self) -> str:
        """Returns the TCL code of the damping"""
        res = f"# damping rayleigh {self.tag} {self.alphaM} {self.betaK} {self.betaKInit} {self.betaKComm}"
        return res
    
    @staticmethod
    def get_Type() -> str:
        return "Rayleigh"


class ModalDamping(DampingBase):
    def __init__(self, **kwargs):
        kwargs = self.validate(**kwargs)
        super().__init__()
        self.numberofModes = kwargs["numberofModes"]
        self.dampingFactors = kwargs["dampingFactors"]
    
    def __str__(self) -> str:
        res = super().__str__()
        res += f"\n\tNumber of Modes: {self.numberofModes}"
        res += f"\n\tDamping Factors: {self.dampingFactors}"
        return res
    
    @staticmethod
    def get_Parameters() -> List[Tuple[str, str]]:
        """Returns the parameters of the damping"""
        return [
            ("numberofModes", "number of modes to consider for modal damping (integer greater than 0)"),
            ("dampingFactors", "damping factors for each mode (list of comma separated floats between 0 and 1)"),
        ]
    
    @staticmethod
    def get_Notes() -> Dict[str, Union[str, list]]:
        """Returns the notes of the damping"""
        return {
        "References": ["https://opensees.github.io/OpenSeesDocumentation/user/manual/model/damping/modalDamping.html",
                       "https://portwooddigital.com/2019/09/12/be-careful-with-modal-damping/",
                       "https://portwooddigital.com/2023/01/25/modal-and-stiffness-proportional-damping/",
                    ]
        }

    @staticmethod
    def validate(**kwargs)-> Dict[str, Union[str, list]]:
        """Validates the damping"""
        numberofModes = 0
        dampingFactors = []

        if "numberofModes" in kwargs:
            numberofModes = kwargs["numberofModes"]
            try :
                numberofModes = int(numberofModes)
            except ValueError:
                raise ValueError("numberofModes should be an integer")
            if numberofModes <= 0:
                raise ValueError("numberofModes should be greater than 0")
        else:
            raise ValueError("numberofModes is required")
        
        if "dampingFactors" in kwargs:
            dampingFactors = kwargs["dampingFactors"]
            try:
                dampingFactors = dampingFactors.split(",")
            except :
                pass
            if not isinstance(dampingFactors, list):
                raise ValueError("dampingFactors should be a list")
            if len(dampingFactors) != numberofModes:
                raise ValueError("dampingFactors should have the same length as numberofModes")
            for i, factor in enumerate(dampingFactors):
                try:
                    dampingFactors[i] = float(factor)
                    if dampingFactors[i] < 0 or dampingFactors[i] > 1:
                        raise ValueError("dampingFactors should be greater than or equal to 0 and less than or equal to 1")
                except ValueError:
                    raise ValueError("dampingFactors should be a list of floats")
        else:
            raise ValueError("dampingFactors is required")
        
        return {
            "numberofModes": numberofModes,
            "dampingFactors": dampingFactors
        }
    
    def get_values(self) -> Dict[str, Union[str, int, float, list]]:
        return {
            "numberofModes": self.numberofModes,
            "dampingFactors": ",".join([str(x) for x in self.dampingFactors])
        }
    
    def update_values(self, **kwargs) -> None:
        """Updates the values of the parameters of the damping"""
        kwargs = self.validate(**kwargs)
        self.numberofModes = kwargs["numberofModes"]
        self.dampingFactors = kwargs["dampingFactors"]
    
    def to_tcl(self) -> str:
        """Returns the TCL code of the damping"""
        res = f"-modalDamping {' '.join([str(x) for x in self.dampingFactors])}"
        return res
    
    @staticmethod
    def get_Type() -> str:
        return "Modal"
                

class FrequencyRayleighDamping(RayleighDamping):
    def __init__(self, **kwargs):
        kwargs = self.validate(**kwargs)
        super().__init__(**kwargs)
        self.f1 = kwargs["f1"]
        self.f2 = kwargs["f2"]
        self.dampingFactor = kwargs["dampingFactor"]
        kwargs = {
            "alphaM": 0.0,
            "betaK": 0.0,
            "betaKInit": 0.0,
            "betaKComm": 0.0,
        }

    def __str__(self) -> str:
        res = super().__str__()
        res += f"\n\tf1: {self.f1}"
        res += f"\n\tf2: {self.f2}"
        res += f"\n\tDamping Factor: {self.dampingFactor}"
        return res
    
    @staticmethod
    def get_Parameters() -> List[Tuple[str, str]]:
        """Returns the parameters of the damping"""
        return [
            ("dampingFactor", "damping factor for the frequency range (float between 0 and 1) (required)"),
            ("f1", "lower bound Target Frequency (float greater than 0) (optional, default=0.2)"),
            ("f2", "upper bound Target Frequency (float greater than 0) (optional, default=20)")
        ]
    
    @staticmethod
    def get_Notes() -> Dict[str, Union[str, list]]:
        """Returns the notes of the damping"""
        return {
        "Notes": [
            '''
                <b>Mathematical Formulation:</b>
                <p style="text-indent: 30px;">ω₁ = 2π f₁,  ω₂ = 2π f₂</p>
                <p style="text-indent: 30px;">α<sub>M</sub> = 2ω₁ω₂ / (ω₁ + ω₂)</p>
                <p style="text-indent: 30px;">β<sub>K</sub> = 2 / (ω₁ + ω₂)</p>
                <p style="text-indent: 30px;">β<sub>KInit</sub> = 0,  β<sub>KComm</sub> = 0</p>
              ''', 
            "The damping matrix is calculated as: C = α<sub>M</sub> M + β<sub>K</sub>  K",
            ],
        "References": []
    }

    @staticmethod
    def validate(**kwargs)-> Dict[str, Union[str, list]]:
        """Validates the damping"""
        f1 = 0.2 # Hz 
        f2 = 20  # Hz
        dampingFactor = 0

        if "f1" in kwargs:
            f1 = kwargs["f1"]
            try :
                f1 = float(f1)
            except ValueError:
                raise ValueError("f1 should be a float")
            if f1 <= 0:
                raise ValueError("f1 should be greater than 0")
        
        if "f2" in kwargs:
            f2 = kwargs["f2"]
            try :
                f2 = float(f2)
            except ValueError:
                raise ValueError("f2 should be a float")
            if f2 <= 0:
                raise ValueError("f2 should be greater than 0")
        
        if "dampingFactor" in kwargs:
            dampingFactor = kwargs["dampingFactor"]
            try:
                dampingFactor = float(dampingFactor)
            except ValueError:
                raise ValueError("dampingFactor should be a float")
            
            if dampingFactor < 0 or dampingFactor > 1:
                raise ValueError("dampingFactor should be greater than or equal to 0 and less than or equal to 1")
        else:
            raise ValueError("dampingFactor is required")
        
        # calculating the damping factors
        omega1 = 2 * 3.141592653589 * f1
        omega2 = 2 * 3.141592653589 * f2
        alphaM = 2 * dampingFactor * omega1 * omega2 / (omega1 + omega2)
        betaK  = (2 * dampingFactor) / (omega1 + omega2)

        return {
            "f1": f1,
            "f2": f2,
            "dampingFactor": dampingFactor,
            "alphaM": alphaM,
            "betaK": betaK,
            "betaKInit": 0,
            "betaKComm": 0
        }
    
    def get_values(self) -> Dict[str, Union[str, int, float, list]]:
        return {
            "dampingFactor": self.dampingFactor,
            "f1": self.f1,
            "f2": self.f2,
        }
    

    def update_values(self, **kwargs) -> None:
        """Updates the values of the parameters of the damping"""
        kwargs = self.validate(**kwargs)
        self.f1 = kwargs["f1"]
        self.f2 = kwargs["f2"]
        self.dampingFactor = kwargs["dampingFactor"]    
        self.alphaM = kwargs["alphaM"]
        self.betaK = kwargs["betaK"]
        self.betaKInit = kwargs["betaKInit"]
        self.betaKComm = kwargs["betaKComm"]

    @staticmethod
    def get_Type() -> str:
        return "Frequency Rayleigh"



class UniformDamping(DampingBase):
    def __init__(self, **kwargs):
        kwargs = self.validate(**kwargs)
        super().__init__()
        self.dampingRatio = kwargs["dampingRatio"]
        self.freql = kwargs["freql"]
        self.freq2 = kwargs["freq2"]
        self.Ta    = kwargs.get("Ta", None)
        self.Td    = kwargs.get("Td", None)
        self.tsTagScaleFactorVsTime = kwargs.get("tsTagScaleFactorVsTime", None)

    def __str__(self) -> str:
        res = super().__str__()
        res += f"\n\tDamping Ratio: {self.dampingRatio}"
        res += f"\n\tLower Frequency: {self.freql}"
        res += f"\n\tUpper Frequency: {self.freq2}"
        res += f"\n\tTa: {self.Ta if self.Ta is not None else 'default'}"
        res += f"\n\tTd: {self.Td if self.Td is not None else 'default'}"
        res += f"\n\ttsTagScaleFactorVsTime: {self.tsTagScaleFactorVsTime if self.tsTagScaleFactorVsTime is not None else 'No time series'}"
        return res
    

    @staticmethod
    def get_Parameters() -> List[Tuple[str, str]]:
        """Returns the parameters of the damping"""
        return [
            ("dampingRatio", "target equivalent viscous damping ratio (float between 0 and 1) (required)"),
            ("freql", "lower bound of the frequency range (in units of T^-1) (float greater than 0) (required)"),
            ("freq2", "upper bound of the frequency range (in units of T^-1) (float greater than 0) (required)"),
            ("Ta", "time when the damping is activated (float)"),
            ("Td", "time when the damping is deactivated (float) (optional)"),
            ("tsTagScaleFactorVsTime", "time series tag identifying the scale factor of the damping versus time (int) (optional)")
        ]
    
    @staticmethod
    def get_Notes() -> Dict[str, Union[str, list]]:
        """Returns the notes of the damping"""
        return {
            "References": [
                "https://opensees.github.io/OpenSeesDocumentation/user/manual/model/damping/elementalDamping/UniformDamping.html",
            ]
        }
    
    @staticmethod
    def validate(**kwargs)-> Dict[str, Union[str, list]]:
        """Validates the damping"""

        # make them float
        try :
            for key in ["dampingRatio", "freql", "freq2", "Ta", "Td"]:
                res = kwargs.get(key, None)
                if res is not None and res != "":
                    kwargs[key] = float(kwargs[key])
        except ValueError:
            raise ValueError(f"{key} should be a float")
        
        # check if the values are between 0 and 1
        for key in ["dampingRatio", "freql", "freq2"]:
            if key not in kwargs:
                raise ValueError(f"{key} is required")
            
        if kwargs["dampingRatio"] < 0 or kwargs["dampingRatio"] > 1:
            raise ValueError("dampingRatio should be greater than or equal to 0 and less than or equal to 1")
        
        if kwargs["freql"] <= 0:
            raise ValueError("freql should be greater than 0")
        
        if kwargs["freq2"] <= 0 or kwargs["freq2"] <= kwargs["freql"]:
            raise ValueError("freq2 should be greater than 0 and greater than freql")
        
        if "tsTagScaleFactorVsTime" in kwargs and kwargs["tsTagScaleFactorVsTime"] != "":
            try:
                kwargs["tsTagScaleFactorVsTime"] = int(kwargs["tsTagScaleFactorVsTime"])
            except ValueError:
                raise ValueError("tsTagScaleFactorVsTime should be an integer")

        
        return kwargs
        
    def get_values(self) -> Dict[str, Union[str, int, float, list]]:
        return {
            "dampingRatio": self.dampingRatio,
            "freql": self.freql,
            "freq2": self.freq2,
            **{k: v for k, v in {
            "Ta": self.Ta,
            "Td": self.Td,
            "tsTagScaleFactorVsTime": self.tsTagScaleFactorVsTime
            }.items() if v is not None}
        }
    
    def update_values(self, **kwargs) -> None:
        """Updates the values of the parameters of the damping"""
        kwargs = self.validate(**kwargs)
        self.dampingRatio = kwargs["dampingRatio"]
        self.freql = kwargs["freql"]
        self.freq2 = kwargs["freq2"]
        
        # Handle optional parameters if they are present
        self.Ta = kwargs.get("Ta", None) # Ta is optional so we use get
        self.Td = kwargs.get("Td", None) # Td is optional so we use get
        self.tsTagScaleFactorVsTime = kwargs.get("tsTagScaleFactorVsTime", None) # tsTagScaleFactorVsTime is optional so we use get

    def to_tcl(self) -> str:
        """Returns the TCL code of the damping"""
        res = f"damping Uniform {self.tag} {self.dampingRatio} {self.freql} {self.freq2}"
        if self.Ta is not None:
            res += f" -activateTime  {self.Ta}"
        if self.Td is not None:
            res += f" -deactivateTime {self.Td}"
        if self.tsTagScaleFactorVsTime is not None:
            res += f" -fact {self.tsTagScaleFactorVsTime}"

        return res
    
    @staticmethod
    def get_Type() -> str:
        return "Uniform"





class SecantStiffnessProportional (DampingBase):
    def __init__(self, **kwargs):
        kwargs = self.validate(**kwargs)
        super().__init__()
        self.dampingFactor = kwargs["dampingFactor"]
        self.Ta = kwargs.get("Ta", None)
        self.Td = kwargs.get("Td", None)
        self.tsTagScaleFactorVsTime = kwargs.get("tsTagScaleFactorVsTime", None)
    
    def __str__(self) -> str:
        res = super().__str__()
        res += f"\n\tDamping Factor: {self.dampingFactor}"
        res += f"\n\tTa: {self.Ta if self.Ta is not None else 'default'}"
        res += f"\n\tTd: {self.Td if self.Td is not None else 'default'}"
        res += f"\n\ttsTagScaleFactorVsTime: {self.tsTagScaleFactorVsTime if self.tsTagScaleFactorVsTime is not None else 'No time series'}"
        return res
    
    @staticmethod
    def get_Parameters() -> List[Tuple[str, str]]:
        """Returns the parameters of the damping"""
        return [
            ("dampingFactor", "coefficient used in the secant stiffness-proportional damping (float between 0 and 1) (required)"),
            ("Ta", "time when the damping is activated (float) (optional)"),
            ("Td", "time when the damping is deactivated (float) (optional)"),
            ("tsTagScaleFactorVsTime", "time series tag identifying the scale factor of the damping versus time (int) (optional)")
        ]
    
    @staticmethod
    def get_Notes() -> Dict[str, Union[str, list]]:
        """Returns the notes of the damping"""
        return {
            "Notes": [
                """The formulation of the damping matrix is:
                <p style="text-indent: 30px;"> f <sub> damping </sub> = dampingFactor * K<sub> secant </sub> * u&#775; </p>
                """,
            ],

            "References": [
                "https://opensees.github.io/OpenSeesDocumentation/user/manual/model/damping/elementalDamping/SecStifDamping.html"
                ]
        }
    
    @staticmethod
    def validate(**kwargs)-> Dict[str, Union[str, list]]:
        """Validates the damping"""
        # make them float
        try :
            for key in ["dampingFactor", "Ta", "Td"]:
                res = kwargs.get(key, None)
                if res is not None and res != "":
                    kwargs[key] = float(kwargs[key])
        except ValueError:
            raise ValueError(f"{key} should be a float")
        
        # check if the values are between 0 and 1
        if "dampingFactor" not in kwargs:
            raise ValueError("dampingFactor is required")
        
        if kwargs["dampingFactor"] < 0 or kwargs["dampingFactor"] > 1:
            raise ValueError("dampingFactor should be greater than or equal to 0 and less than or equal to 1")
        
        if "tsTagScaleFactorVsTime" in kwargs and kwargs["tsTagScaleFactorVsTime"] != "":
            try:
                kwargs["tsTagScaleFactorVsTime"] = int(kwargs["tsTagScaleFactorVsTime"])
            except ValueError:
                raise ValueError("tsTagScaleFactorVsTime should be an integer")

        return kwargs
    
    def get_values(self) -> Dict[str, Union[str, int, float, list]]:
        return {
            "dampingFactor": self.dampingFactor,
            **{k: v for k, v in {
            "Ta": self.Ta,
            "Td": self.Td,
            "tsTagScaleFactorVsTime": self.tsTagScaleFactorVsTime
            }.items() if v is not None}
        }
    
    def update_values(self, **kwargs) -> None:
        """Updates the values of the parameters of the damping"""
        kwargs = self.validate(**kwargs)
        self.dampingFactor = kwargs["dampingFactor"]

        # Handle optional parameters if they are present
        self.Ta = kwargs.get("Ta", None)
        self.Td = kwargs.get("Td", None)
        self.tsTagScaleFactorVsTime = kwargs.get("tsTagScaleFactorVsTime", None)
    
    def to_tcl(self) -> str:
        """Returns the TCL code of the damping"""
        res = f"damping SecStiff {self.tag} {self.dampingFactor}"
        if self.Ta is not None:
            res += f" -activateTime  {self.Ta}"
        if self.Td is not None:
            res += f" -deactivateTime {self.Td}"
        if self.tsTagScaleFactorVsTime is not None:
            res += f" -fact {self.tsTagScaleFactorVsTime}"

        return res
    
    @staticmethod
    def get_Type() -> str:
        return "Secant Stiffness Proportional"



class DampingManager:
    """
    A centralized manager for all damping instances in MeshMaker.
    
    The DampingManager implements the Singleton pattern to ensure a single, consistent
    point of damping management across the entire application. It provides methods for
    creating, retrieving, and managing damping objects used in dynamic structural analysis.
    
    All damping objects created through this manager are automatically tracked and tagged,
    simplifying the process of creating models with appropriate energy dissipation mechanisms.
    
    Usage:
        # Direct access
        from femora.components.Damping import DampingManager
        damping_manager = DampingManager()
        
        # Through MeshMaker (recommended)
        from femora.components.MeshMaker import MeshMaker
        mk = MeshMaker()
        damping_manager = mk.damping
        
        # Creating a damping instance
        rayleigh_damping = damping_manager.create_damping('rayleigh', alphaM=0.05, betaK=0.001)
    """
    _instance = None
    
    def __new__(cls):
        """
        Create a new DampingManager instance or return the existing one if already created.
        
        Returns:
            DampingManager: The singleton instance of the DampingManager
        """
        if cls._instance is None:
            cls._instance = super(DampingManager, cls).__new__(cls)
        return cls._instance

    @classmethod
    def get_instance(cls):
        """
        Get the singleton instance of the DampingManager.
        
        This method ensures that only one instance of DampingManager exists throughout
        the application, following the Singleton design pattern.
        
        Returns:
            DampingManager: The singleton instance of the DampingManager
        """
        return cls._instance

    def create_damping(self, damping_type: str, **kwargs) -> DampingBase:
        """
        Create a new damping instance of the specified type.
        
        This method delegates the damping creation to the DampingRegistry, which
        maintains a dictionary of available damping types and their corresponding classes.
        
        Args:
            damping_type (str): The type of damping to create. Available types include:
                - 'rayleigh': Classical Rayleigh damping
                - 'modal': Modal damping for specific modes
                - 'frequency rayleigh': Rayleigh damping specified by target frequencies
                - 'uniform': Uniform damping across a frequency range
                - 'secant stiffness proportional': Damping proportional to secant stiffness
            **kwargs: Specific parameters for the damping type being created
                      (see documentation for each damping type for details)
        
        Returns:
            DampingBase: A new instance of the requested damping type
            
        Raises:
            ValueError: If the damping type is unknown or if required parameters are missing or invalid
        """
        return DampingRegistry.create_damping(damping_type, **kwargs)

    def get_damping(self, tag: int) -> DampingBase:
        """
        Get a damping instance by its tag.
        
        Args:
            tag (int): The unique identifier of the damping instance
            
        Returns:
            DampingBase: The damping instance with the specified tag
            
        Raises:
            KeyError: If no damping with the given tag exists
        """
        return DampingBase.get_damping(tag)

    def remove_damping(self, tag: int) -> None:
        """
        Remove a damping instance by its tag.
        
        This method removes the specified damping instance and automatically
        updates the tags of the remaining damping instances to maintain sequential numbering.
        
        Args:
            tag (int): The unique identifier of the damping instance to remove
        """
        DampingBase.remove_damping(tag)

    def get_all_dampings(self) -> Dict[int, DampingBase]:
        """
        Get all damping instances currently managed by this DampingManager.
        
        Returns:
            Dict[int, DampingBase]: A dictionary mapping tags to damping instances
        """
        return DampingBase._dampings

    def clear_all_dampings(self) -> None:
        """
        Remove all damping instances managed by this DampingManager.
        
        This method clears all damping instances from the manager, effectively
        resetting the damping system.
        """
        DampingBase._dampings.clear()

    def get_available_types(self) -> List[str]:
        """
        Get a list of available damping types that can be created.
        
        Returns:
            List[str]: A list of damping type names that can be used with create_damping()
        """
        return DampingRegistry.get_available_types()
    


class DampingRegistry:
    _damping_types = {
        'frequency rayleigh': FrequencyRayleighDamping,
        'rayleigh': RayleighDamping,
        'modal': ModalDamping,
        'uniform': UniformDamping,
        'secant stiffness proportional': SecantStiffnessProportional
    }

    @classmethod
    def create_damping(cls, damping_type: str, **kwargs) -> DampingBase:
        if damping_type.lower() not in cls._damping_types:
            raise ValueError(f"Unknown damping type: {damping_type}")
        return cls._damping_types[damping_type.lower()](**kwargs)

    @classmethod
    def get_available_types(cls) -> List[str]:
        return list(cls._damping_types.keys())

    @classmethod
    def register_damping_type(cls, name: str, damping_class: Type[DampingBase]):
        cls._damping_types[name.lower()] = damping_class

    @classmethod
    def remove_damping_type(cls, name: str):
        if name.lower() in cls._damping_types:
            del cls._damping_types[name.lower()]

 



if __name__ == "__main__":
    # test the DampingBase class
    damping1 = DampingBase()
    damping2 = DampingBase()
    damping3 = DampingBase()

    RayleighDamping1 = RayleighDamping(alphaM=0.1, betaK=0.2, betaKInit=0.3, betaKComm=0.4)
    RayleighDamping2 = RayleighDamping(alphaM=0.5, betaK=0.6, betaKInit=0.7, betaKComm=0.8)
    ModalDamping1 = ModalDamping(numberofModes=2, dampingFactors="0.1,0.2")
    ModalDamping2 = ModalDamping(numberofModes=3, dampingFactors="0.3,0.4,0.5")



    DampingBase.print_dampings()

    print(10*"*"+"\nRemoving damping 2\n"+10*"*")

    DampingBase.remove_damping(2)
    DampingBase.print_dampings()




from qtpy.QtGui import QDoubleValidator, QIntValidator



class DoubleValidator(QDoubleValidator):
    def validate(self, input_string, pos):
        if ',' in input_string:
            return (QDoubleValidator.Invalid, input_string, pos)
        return super().validate(input_string, pos)
    
class IntValidator(QIntValidator):
    def validate(self, input_string, pos):
        if ',' in input_string or input_string.startswith('0'):
            return (QIntValidator.Invalid, input_string, pos)
        return super().validate(input_string, pos)
    
class PositiveDoubleValidator(QDoubleValidator):
    def validate(self, input_string, pos):
        if ',' in input_string or float(input_string) <= 0:
            return (QDoubleValidator.Invalid, input_string, pos)
        return super().validate(input_string, pos)


from typing import List, Dict, Optional, Union, Type
from .base import AnalysisComponent
from abc import abstractmethod

class Algorithm(AnalysisComponent):
    """
    Base abstract class for algorithms, which determine how the constraint equations are enforced
    in the system of equations
    """
    _algorithms = {}  # Class-level dictionary to store algorithm types
    _created_algorithms = {}  # Class-level dictionary to track all created algorithms
    _next_tag = 1  # Class variable to track the next tag to assign
    
    def __init__(self, algorithm_type: str):
        """
        Initialize an algorithm
        
        Args:
            algorithm_type (str): Type of the algorithm
        """
        self.tag = Algorithm._next_tag
        Algorithm._next_tag += 1
        self.algorithm_type = algorithm_type
        
        # Register this algorithm in the class-level tracking dictionary
        Algorithm._created_algorithms[self.tag] = self
    
    @staticmethod
    def register_algorithm(name: str, algorithm_class: Type['Algorithm']):
        """Register an algorithm type"""
        Algorithm._algorithms[name.lower()] = algorithm_class
    
    @staticmethod
    def create_algorithm(algorithm_type: str, **kwargs) -> 'Algorithm':
        """Create an algorithm of the specified type"""
        algorithm_type = algorithm_type.lower()
        if algorithm_type not in Algorithm._algorithms:
            raise ValueError(f"Unknown algorithm type: {algorithm_type}")
        return Algorithm._algorithms[algorithm_type](**kwargs)
    
    @staticmethod
    def get_available_types() -> List[str]:
        """Get available algorithm types"""
        return list(Algorithm._algorithms.keys())
    
    @classmethod
    def get_algorithm(cls, tag: int) -> 'Algorithm':
        """
        Retrieve a specific algorithm by its tag.
        
        Args:
            tag (int): The tag of the algorithm
        
        Returns:
            Algorithm: The algorithm with the specified tag
        
        Raises:
            KeyError: If no algorithm with the given tag exists
        """
        if tag not in cls._created_algorithms:
            raise KeyError(f"No algorithm found with tag {tag}")
        return cls._created_algorithms[tag]

    @classmethod
    def get_all_algorithms(cls) -> Dict[int, 'Algorithm']:
        """
        Retrieve all created algorithms.
        
        Returns:
            Dict[int, Algorithm]: A dictionary of all algorithms, keyed by their unique tags
        """
        return cls._created_algorithms
    
    @classmethod
    def clear_all(cls) -> None:
        """
        Clear all algorithms and reset tags.
        """
        cls._created_algorithms.clear()
        cls._next_tag = 1
    
    @abstractmethod
    def get_values(self) -> Dict[str, Union[str, int, float, bool]]:
        """
        Get the parameters defining this algorithm
        
        Returns:
            Dict[str, Union[str, int, float, bool]]: Dictionary of parameter values
        """
        pass

    @classmethod
    def _reassign_tags(cls) -> None:
        """
        Reassign tags to all algorithms sequentially starting from 1.
        """
        new_algorithms = {}
        for idx, algorithm in enumerate(sorted(cls._created_algorithms.values(), key=lambda h: h.tag), start=1):
            algorithm.tag = idx
            new_algorithms[idx] = algorithm
        cls._created_algorithms = new_algorithms
        cls._next_tag = len(cls._created_algorithms) + 1

    @classmethod
    def remove_algorithm(cls, tag: int) -> None:
        """
        Delete an algorithm by its tag and re-tag all remaining algorithms sequentially.
        
        Args:
            tag (int): The tag of the algorithm to delete
        """
        if tag in cls._created_algorithms:
            del cls._created_algorithms[tag]
            cls._reassign_tags()


class LinearAlgorithm(Algorithm):
    """
    Linear algorithm which takes one iteration to solve the system of equations
    """
    def __init__(self, initial: bool = False, factor_once: bool = False):
        """
        Initialize a Linear algorithm
        
        Args:
            initial (bool): Flag to use initial stiffness
            factor_once (bool): Flag to only set up and factor matrix once
        """
        super().__init__("Linear")
        self.initial = initial
        self.factor_once = factor_once
    
    def to_tcl(self) -> str:
        """
        Convert the algorithm to a TCL command string for OpenSees
        
        Returns:
            str: The TCL command string
        """
        cmd = "algorithm Linear"
        if self.initial:
            cmd += " -initial"
        if self.factor_once:
            cmd += " -factorOnce"
        return cmd
    
    def get_values(self) -> Dict[str, Union[str, int, float, bool]]:
        """
        Get the parameters defining this algorithm
        
        Returns:
            Dict[str, Union[str, int, float, bool]]: Dictionary of parameter values
        """
        return {
            "initial": self.initial,
            "factor_once": self.factor_once
        }


class NewtonAlgorithm(Algorithm):
    """
    Newton algorithm uses the Newton-Raphson method to solve the nonlinear residual equation
    """
    def __init__(self, initial: bool = False, initial_then_current: bool = False):
        """
        Initialize a Newton algorithm
        
        Args:
            initial (bool): Flag to use initial stiffness
            initial_then_current (bool): Flag to use initial stiffness on first step and then current stiffness
        """
        super().__init__("Newton")
        self.initial = initial
        self.initial_then_current = initial_then_current
        
        # Check for incompatible options
        if self.initial and self.initial_then_current:
            raise ValueError("Cannot specify both -initial and -initialThenCurrent flags")
    
    def to_tcl(self) -> str:
        """
        Convert the algorithm to a TCL command string for OpenSees
        
        Returns:
            str: The TCL command string
        """
        cmd = "algorithm Newton"
        if self.initial:
            cmd += " -initial"
        if self.initial_then_current:
            cmd += " -initialThenCurrent"
        return cmd
    
    def get_values(self) -> Dict[str, Union[str, int, float, bool]]:
        """
        Get the parameters defining this algorithm
        
        Returns:
            Dict[str, Union[str, int, float, bool]]: Dictionary of parameter values
        """
        return {
            "initial": self.initial,
            "initial_then_current": self.initial_then_current
        }


class ModifiedNewtonAlgorithm(Algorithm):
    """
    Modified Newton algorithm uses the modified Newton-Raphson algorithm to solve the nonlinear residual equation
    """
    def __init__(self, initial: bool = False, factor_once: bool = False):
        """
        Initialize a Modified Newton algorithm
        
        Args:
            initial (bool): Flag to use initial stiffness iterations
            factor_once (bool): Flag to factor matrix only once
        """
        super().__init__("ModifiedNewton")
        self.initial = initial
        self.factor_once = factor_once
    
    def to_tcl(self) -> str:
        """
        Convert the algorithm to a TCL command string for OpenSees
        
        Returns:
            str: The TCL command string
        """
        cmd = "algorithm ModifiedNewton"
        if self.initial:
            cmd += " -initial"
        if self.factor_once:
            cmd += " -factoronce"
        return cmd
    
    def get_values(self) -> Dict[str, Union[str, int, float, bool]]:
        """
        Get the parameters defining this algorithm
        
        Returns:
            Dict[str, Union[str, int, float, bool]]: Dictionary of parameter values
        """
        return {
            "initial": self.initial,
            "factor_once": self.factor_once
        }


class NewtonLineSearchAlgorithm(Algorithm):
    """
    Newton Line Search algorithm introduces line search to the Newton-Raphson algorithm
    """
    def __init__(self, type_search: str = "InitialInterpolated", tol: float = 0.8, 
                 max_iter: int = 10, min_eta: float = 0.1, max_eta: float = 10.0):
        """
        Initialize a Newton Line Search algorithm
        
        Args:
            type_search (str): Line search algorithm type
            tol (float): Tolerance for search
            max_iter (int): Maximum number of iterations to try
            min_eta (float): Minimum η value
            max_eta (float): Maximum η value
        """
        super().__init__("NewtonLineSearch")
        self.type_search = type_search
        self.tol = tol
        self.max_iter = max_iter
        self.min_eta = min_eta
        self.max_eta = max_eta
        
        # Validate search type
        valid_search_types = ["Bisection", "Secant", "RegulaFalsi", "InitialInterpolated"]
        if self.type_search not in valid_search_types:
            raise ValueError(f"Invalid search type: {self.type_search}. "
                           f"Valid types are: {', '.join(valid_search_types)}")
    
    def to_tcl(self) -> str:
        """
        Convert the algorithm to a TCL command string for OpenSees
        
        Returns:
            str: The TCL command string
        """
        cmd = f"algorithm NewtonLineSearch -type {self.type_search}"
        
        # Add other parameters if they're not default values
        if self.tol != 0.8:
            cmd += f" -tol {self.tol}"
        if self.max_iter != 10:
            cmd += f" -maxIter {self.max_iter}"
        if self.min_eta != 0.1:
            cmd += f" -minEta {self.min_eta}"
        if self.max_eta != 10.0:
            cmd += f" -maxEta {self.max_eta}"
            
        return cmd
    
    def get_values(self) -> Dict[str, Union[str, int, float, bool]]:
        """
        Get the parameters defining this algorithm
        
        Returns:
            Dict[str, Union[str, int, float, bool]]: Dictionary of parameter values
        """
        return {
            "type_search": self.type_search,
            "tol": self.tol,
            "max_iter": self.max_iter,
            "min_eta": self.min_eta,
            "max_eta": self.max_eta
        }


class KrylovNewtonAlgorithm(Algorithm):
    """
    Krylov-Newton algorithm uses a modified Newton method with Krylov subspace acceleration
    """
    def __init__(self, tang_iter: str = "current", tang_incr: str = "current", max_dim: int = 3):
        """
        Initialize a Krylov-Newton algorithm
        
        Args:
            tang_iter (str): Tangent to iterate on
            tang_incr (str): Tangent to increment on
            max_dim (int): Max number of iterations until tangent is reformed
        """
        super().__init__("KrylovNewton")
        self.tang_iter = tang_iter
        self.tang_incr = tang_incr
        self.max_dim = max_dim
        
        # Validate tangent options
        valid_tangent_options = ["current", "initial", "noTangent"]
        if self.tang_iter not in valid_tangent_options:
            raise ValueError(f"Invalid tangent iteration type: {self.tang_iter}. "
                           f"Valid types are: {', '.join(valid_tangent_options)}")
        if self.tang_incr not in valid_tangent_options:
            raise ValueError(f"Invalid tangent increment type: {self.tang_incr}. "
                           f"Valid types are: {', '.join(valid_tangent_options)}")
    
    def to_tcl(self) -> str:
        """
        Convert the algorithm to a TCL command string for OpenSees
        
        Returns:
            str: The TCL command string
        """
        cmd = "algorithm KrylovNewton"
        
        # Add parameters if they're not default values
        if self.tang_iter != "current":
            cmd += f" -iterate {self.tang_iter}"
        if self.tang_incr != "current":
            cmd += f" -increment {self.tang_incr}"
        if self.max_dim != 3:
            cmd += f" -maxDim {self.max_dim}"
            
        return cmd
    
    def get_values(self) -> Dict[str, Union[str, int, float, bool]]:
        """
        Get the parameters defining this algorithm
        
        Returns:
            Dict[str, Union[str, int, float, bool]]: Dictionary of parameter values
        """
        return {
            "tang_iter": self.tang_iter,
            "tang_incr": self.tang_incr,
            "max_dim": self.max_dim
        }


class SecantNewtonAlgorithm(Algorithm):
    """
    Secant Newton algorithm uses the two-term update to accelerate convergence
    """
    def __init__(self, tang_iter: str = "current", tang_incr: str = "current", max_dim: int = 3):
        """
        Initialize a Secant Newton algorithm
        
        Args:
            tang_iter (str): Tangent to iterate on
            tang_incr (str): Tangent to increment on
            max_dim (int): Max number of iterations until tangent is reformed
        """
        super().__init__("SecantNewton")
        self.tang_iter = tang_iter
        self.tang_incr = tang_incr
        self.max_dim = max_dim
        
        # Validate tangent options
        valid_tangent_options = ["current", "initial", "noTangent"]
        if self.tang_iter not in valid_tangent_options:
            raise ValueError(f"Invalid tangent iteration type: {self.tang_iter}. "
                           f"Valid types are: {', '.join(valid_tangent_options)}")
        if self.tang_incr not in valid_tangent_options:
            raise ValueError(f"Invalid tangent increment type: {self.tang_incr}. "
                           f"Valid types are: {', '.join(valid_tangent_options)}")
    
    def to_tcl(self) -> str:
        """
        Convert the algorithm to a TCL command string for OpenSees
        
        Returns:
            str: The TCL command string
        """
        cmd = "algorithm SecantNewton"
        
        # Add parameters if they're not default values
        if self.tang_iter != "current":
            cmd += f" -iterate {self.tang_iter}"
        if self.tang_incr != "current":
            cmd += f" -increment {self.tang_incr}"
        if self.max_dim != 3:
            cmd += f" -maxDim {self.max_dim}"
            
        return cmd
    
    def get_values(self) -> Dict[str, Union[str, int, float, bool]]:
        """
        Get the parameters defining this algorithm
        
        Returns:
            Dict[str, Union[str, int, float, bool]]: Dictionary of parameter values
        """
        return {
            "tang_iter": self.tang_iter,
            "tang_incr": self.tang_incr,
            "max_dim": self.max_dim
        }


class BFGSAlgorithm(Algorithm):
    """
    BFGS algorithm performs successive rank-two updates of the tangent (for symmetric systems)
    """
    def __init__(self, count: int):
        """
        Initialize a BFGS algorithm
        
        Args:
            count (int): Number of iterations before reforming tangent
        """
        super().__init__("BFGS")
        self.count = count
        
        # Validate count
        if not isinstance(self.count, int) or self.count < 1:
            raise ValueError("Count must be a positive integer")
    
    def to_tcl(self) -> str:
        """
        Convert the algorithm to a TCL command string for OpenSees
        
        Returns:
            str: The TCL command string
        """
        return f"algorithm BFGS {self.count}"
    
    def get_values(self) -> Dict[str, Union[str, int, float, bool]]:
        """
        Get the parameters defining this algorithm
        
        Returns:
            Dict[str, Union[str, int, float, bool]]: Dictionary of parameter values
        """
        return {
            "count": self.count
        }


class BroydenAlgorithm(Algorithm):
    """
    Broyden algorithm performs successive rank-one updates of the tangent (for general unsymmetric systems)
    """
    def __init__(self, count: int):
        """
        Initialize a Broyden algorithm
        
        Args:
            count (int): Number of iterations before reforming tangent
        """
        super().__init__("Broyden")
        self.count = count
        
        # Validate count
        if not isinstance(self.count, int) or self.count < 1:
            raise ValueError("Count must be a positive integer")
    
    def to_tcl(self) -> str:
        """
        Convert the algorithm to a TCL command string for OpenSees
        
        Returns:
            str: The TCL command string
        """
        return f"algorithm Broyden {self.count}"
    
    def get_values(self) -> Dict[str, Union[str, int, float, bool]]:
        """
        Get the parameters defining this algorithm
        
        Returns:
            Dict[str, Union[str, int, float, bool]]: Dictionary of parameter values
        """
        return {
            "count": self.count
        }


class ExpressNewtonAlgorithm(Algorithm):
    """
    ExpressNewton algorithm accepts the solution after a constant number of iterations
    """
    def __init__(self, iter_count: int = 2, k_multiplier: float = 1.0, 
                 initial_tangent: bool = False, current_tangent: bool = True, 
                 factor_once: bool = False):
        """
        Initialize an ExpressNewton algorithm
        
        Args:
            iter_count (int): Constant number of iterations
            k_multiplier (float): Multiplier to system stiffness
            initial_tangent (bool): Flag to use initial stiffness
            current_tangent (bool): Flag to use current stiffness
            factor_once (bool): Flag to factorize system matrix only once
        """
        super().__init__("ExpressNewton")
        self.iter_count = iter_count
        self.k_multiplier = k_multiplier
        self.initial_tangent = initial_tangent
        self.current_tangent = current_tangent
        self.factor_once = factor_once
        
        # Validate iter_count
        if not isinstance(self.iter_count, int) or self.iter_count < 1:
            raise ValueError("Iteration count must be a positive integer")
        
        # Check for incompatible options
        if self.initial_tangent and self.current_tangent:
            raise ValueError("Cannot specify both -initialTangent and -currentTangent flags")
    
    def to_tcl(self) -> str:
        """
        Convert the algorithm to a TCL command string for OpenSees
        
        Returns:
            str: The TCL command string
        """
        cmd = f"algorithm ExpressNewton {self.iter_count} {self.k_multiplier}"
        
        # Add optional flags
        if self.initial_tangent:
            cmd += " -initialTangent"
        if self.current_tangent and not self.initial_tangent:
            cmd += " -currentTangent"
        if self.factor_once:
            cmd += " -factorOnce"
            
        return cmd
    
    def get_values(self) -> Dict[str, Union[str, int, float, bool]]:
        """
        Get the parameters defining this algorithm
        
        Returns:
            Dict[str, Union[str, int, float, bool]]: Dictionary of parameter values
        """
        return {
            "iter_count": self.iter_count,
            "k_multiplier": self.k_multiplier,
            "initial_tangent": self.initial_tangent,
            "current_tangent": self.current_tangent,
            "factor_once": self.factor_once
        }


class AlgorithmManager:
    """
    Singleton class for managing algorithms
    """
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(AlgorithmManager, cls).__new__(cls)
        return cls._instance

    def create_algorithm(self, algorithm_type: str, **kwargs) -> Algorithm:
        """Create a new algorithm"""
        return Algorithm.create_algorithm(algorithm_type, **kwargs)

    def get_algorithm(self, tag: int) -> Algorithm:
        """Get algorithm by tag"""
        return Algorithm.get_algorithm(tag)

    def remove_algorithm(self, tag: int) -> None:
        """Remove algorithm by tag"""
        Algorithm.remove_algorithm(tag)

    def get_all_algorithms(self) -> Dict[int, Algorithm]:
        """Get all algorithms"""
        return Algorithm.get_all_algorithms()

    def get_available_types(self) -> List[str]:
        """Get list of available algorithm types"""
        return Algorithm.get_available_types()
    
    def clear_all(self):
        """Clear all algorithms"""  
        Algorithm.clear_all()


# Register all algorithms
Algorithm.register_algorithm('linear', LinearAlgorithm)
Algorithm.register_algorithm('newton', NewtonAlgorithm)
Algorithm.register_algorithm('modifiednewton', ModifiedNewtonAlgorithm)
Algorithm.register_algorithm('newtonlinesearch', NewtonLineSearchAlgorithm)
Algorithm.register_algorithm('krylovnewton', KrylovNewtonAlgorithm)
Algorithm.register_algorithm('secantnewton', SecantNewtonAlgorithm)
Algorithm.register_algorithm('bfgs', BFGSAlgorithm)
Algorithm.register_algorithm('broyden', BroydenAlgorithm)
Algorithm.register_algorithm('expressnewton', ExpressNewtonAlgorithm)


from typing import List, Dict
from abc import ABC, abstractmethod
from femora.components.Assemble.Assembler import Assembler
from pykdtree.kdtree import KDTree
import numpy as np

class mpConstraint(ABC):
    """Base class for OpenSees multi-point constraints"""
    
    # Class variable to store all MP constraints
    _constraints: Dict[int, 'mpConstraint'] = {}
    

    def __init__(self, master_node: int, slave_nodes: List[int]):
        """
        Initialize the base MP constraint
        
        Args:
            master_node: Tag of the master/retained node
            slave_nodes: List of slave/constrained node tags
        """
        self.master_node = master_node
        self.slave_nodes = slave_nodes
        self.tag = mpConstraint._next_tag()
        mpConstraint._constraints[self.tag] = self
        
    
    @classmethod
    def _next_tag(cls) -> int:
        """Get the next available tag"""
        return len(cls._constraints) + 1
    
    @classmethod
    def get_constraint(cls, tag: int) -> 'mpConstraint':
        """Get a constraint by its tag"""
        return cls._constraints.get(tag)
    
    @classmethod
    def remove_constraint(cls, tag: int) -> None:
        """Remove a constraint and reorder remaining tags"""
        if tag in cls._constraints:
            del cls._constraints[tag]
            # Reorder remaining constraints
            constraints = sorted(cls._constraints.items())
            cls._constraints.clear()
            for new_tag, (_, constraint) in enumerate(constraints, 1):
                constraint.tag = new_tag
                cls._constraints[new_tag] = constraint
    
    @abstractmethod
    def to_tcl(self) -> str:
        """Convert constraint to TCL command for OpenSees"""
        pass





class equalDOF(mpConstraint):
    """Equal DOF constraint"""
    
    def __init__(self, master_node: int, slave_nodes: List[int], dofs: List[int]):
        """
        Initialize EqualDOF constraint
        
        Args:
            master_node: Retained node tag
            slave_node: Constrained node tag
            dofs: List of DOFs to be constrained
        """
        super().__init__(master_node, slave_nodes)
        self.dofs = dofs
    
    def to_tcl(self) -> str:
        tcl_str = ""
        for slave in self.slave_nodes:
            dofs_str = " ".join(str(dof) for dof in self.dofs)
            tcl_str += f"equalDOF {self.master_node} {slave} {dofs_str}"
        return tcl_str







class rigidLink(mpConstraint):
    """Rigid link constraint"""
    
    def __init__(self, type_str: str, master_node: int, slave_nodes: List[int]):
        """
        Initialize RigidLink constraint
        
        Args:
            type_str: Type of rigid link ('bar' or 'beam')
            master_node: Retained node tag
            slave_node: Constrained node tag
        """
        super().__init__(master_node, slave_nodes)
        if type_str not in ['bar', 'beam']:
            raise ValueError("RigidLink type must be 'bar' or 'beam'")
        self.type = type_str
    
    def to_tcl(self) -> str:
        tcl_str = ""
        for slave in self.slave_nodes:
            tcl_str += f"rigidLink {self.type} {self.master_node} {slave}"
        # return f"rigidLink {self.type} {self.master_node} {self.slave_nodes[0]}"






class rigidDiaphragm(mpConstraint):
    """Rigid diaphragm constraint"""
    
    def __init__(self, direction: int, master_node: int, slave_nodes: List[int]):
        """
        Initialize RigidDiaphragm constraint
        
        Args:
            direction: Direction perpendicular to rigid plane (3D) or direction of motion (2D)
            master_node: Retained node tag
            slave_nodes: List of constrained node tags
        """
        super().__init__(master_node, slave_nodes)
        self.direction = direction
    
    def to_tcl(self) -> str:
        slaves_str = " ".join(str(node) for node in self.slave_nodes)
        return f"rigidDiaphragm {self.direction} {self.master_node} {slaves_str}"
    






class mpConstraintManager:
    """
    Singleton class to manage MP constraints.
    This class provides methods to create and manage different types of constraints
    such as EqualDOF, RigidLink, and RigidDiaphragm. It ensures that only one instance
    of the class exists and provides a global point of access to it.
    """   
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(mpConstraintManager, cls).__new__(cls)
        return cls._instance

    def create_equal_dof(self, master_node: int, slave_nodes: List[int], dofs: List[int]) -> equalDOF:
        """
        Create an EqualDOF constraint.

        Parameters:
        master_node (int): The master node ID.
        slave_nodes (List[int]): A list of slave node IDs.
        dofs (List[int]): A list of degrees of freedom to be constrained.

        Returns:
        equalDOF: An instance of the equalDOF constraint.
        """
        """Create an EqualDOF constraint"""
        return equalDOF(master_node, slave_nodes, dofs)


    def create_rigid_link(self, type_str: str, master_node: int, slave_nodes: List[int]) -> rigidLink:
        """
        Create a RigidLink constraint.

        Args:
            type_str (str): The type of the rigid link.
            master_node (int): The master node ID.
            slave_nodes (List[int]): A list of slave node IDs.

        Returns:
            rigidLink: An instance of the rigidLink class representing the constraint.
        """
        return rigidLink(type_str, master_node, slave_nodes)


    def create_rigid_diaphragm(self, direction: int, master_node: int, slave_nodes: List[int]) -> rigidDiaphragm:
        """
        Create a RigidDiaphragm constraint.

        Parameters:
        direction (int): The direction of the rigid diaphragm constraint.
        master_node (int): The master node ID for the rigid diaphragm.
        slave_nodes (List[int]): A list of slave node IDs that will be constrained by the rigid diaphragm.

        Returns:
        rigidDiaphragm: An instance of the rigidDiaphragm class representing the created constraint.
        """
        return rigidDiaphragm(direction, master_node, slave_nodes)


    def create_laminar_boundary(self, dofs: List[int], direction: int = 3, tol=1e-2) -> None:
        """
        Create a laminar boundary constraint.

        Parameters:
            dofs (List[int]): A list of degrees of freedom to be constrained.
            direction (int): The direction of the perpendicular plane that will be used to create the laminar boundary.
            tol (float): The tolerance value for the laminar boundary condition.

        Note:
            The direction must be 1, 2, or 3.
            This function will check if the mesh is 3D, and the boundaries will be box-shaped or cylinder-shaped.

        Returns:
            None
        """
        assembler = Assembler()
        # check if the AssembeledMesh is not None
        if assembler.AssembeledMesh is None:
            raise ValueError("AssembeledMesh is not created yet")

        mesh = assembler.AssembeledMesh.copy()
        surface = mesh.extract_surface()
        xmin, xmax, ymin, ymax, zmin, zmax = surface.bounds
        eps = tol
        if direction == 3:
            xmin += eps
            xmax -= eps
            ymin += eps
            ymax -= eps
            zmin -= eps
            zmax += eps
        elif direction == 2:
            xmin += eps
            xmax -= eps
            ymin -= eps
            ymax += eps
            zmin += eps
            zmax -= eps
        elif direction == 1:
            xmin -= eps
            xmax += eps
            ymin += eps
            ymax -= eps
            zmin += eps
            zmax -= eps
        else:
            raise ValueError("Direction must be 1, 2, or 3")
        

        ind = surface.find_cells_within_bounds((xmin, xmax, ymin, ymax, zmin, zmax))
        surface = surface.extract_cells(ind,invert=True)
        points = surface.points
        tolerance = tol

        kdtree = KDTree(mesh.points)
        distances, indices = kdtree.query(points, k=1)
        nodeTags = indices + 1
        
        import pyvista as pv
        # attach the nodeTags to the points at the first column
        z = points[:, direction-1]
        z_rounded = np.round(z / tolerance) * tolerance

        # Use np.unique with return_inverse to group fast
        unique_z, group_ids = np.unique(z_rounded, return_inverse=True)
    

        # Build dictionary-like grouping in NumPy style
        Taggroups = [nodeTags[group_ids == i] for i in range(len(unique_z))]
        groups = [points[group_ids == i] for i in range(len(unique_z))]


        # now add mp constraints
        print(f"Number of groups: {len(Taggroups)}")
        print(f"Number of groups: {len(groups)}")
        for i, group in enumerate(Taggroups):
            if len(group) > 1:
                # self.create_equal_dof(master_node=group[0], slave_nodes=group[1:], dofs=dofs)
                for slave in group[1:]:
                    self.create_equal_dof(master_node=group[0], slave_nodes=[slave], dofs=dofs)
    


    def create_constraint(self, constraint_type: str, *args) -> mpConstraint:
        """
        Create a constraint based on the specified type.

        Parameters:
            constraint_type (str): The type of constraint to create. Supported types are "equalDOF", "rigidLink", and "rigidDiaphragm".
            *args: Additional arguments required for creating the specific type of constraint.

        Returns:
            mpConstraint: An instance of the created constraint.

        Raises:
            ValueError: If an unknown constraint type is provided.
        """
        if constraint_type == "equalDOF":
            return self.create_equal_dof(*args)
        elif constraint_type == "rigidLink":
            return self.create_rigid_link(*args)
        elif constraint_type == "rigidDiaphragm":
            return self.create_rigid_diaphragm(*args)
        else:
            raise ValueError(f"Unknown constraint type: {constraint_type}")


    def get_constraint(self, tag: int) -> mpConstraint:
        """
        Retrieve a constraint by its tag.

        Args:
            tag (int): The tag identifier of the constraint.

        Returns:
            mpConstraint: The constraint object associated with the given tag.
        """
        """Get a constraint by its tag"""
        return mpConstraint.get_constraint(tag)


    def remove_constraint(self, tag: int) -> None:
        """
        Remove a constraint by its tag.

        Parameters:
        tag (int): The tag of the constraint to be removed.

        Returns:
        None
        """
        """Remove a constraint by its tag"""
        mpConstraint.remove_constraint(tag)

    
    def __iter__(self):
        """
        Iterate over all constraints.

        This method allows the mpConstraintManager to be iterable, yielding each
        constraint stored in the class variable `_constraints` of `mpConstraint`.

        Yields:
            mpConstraint: Each constraint object in the order of their tags.
        """
        return iter(mpConstraint._constraints.values())




    def to_tcl(self) -> str:
        """
        Convert all constraints to TCL commands.

        This method iterates over all constraints stored in the class variable
        `_constraints` of `mpConstraint` and converts each constraint to its
        corresponding TCL command string. The resulting TCL command strings are
        concatenated into a single string.

        Returns:
            str: A string containing all the TCL commands for the constraints.
        """
        """Convert all constraints to TCL commands"""
        tcl_str = ""
        for constraint in mpConstraint._constraints.values():
            tcl_str += constraint.to_tcl()
        return tcl_str


from femora.components.Damping.dampingBase import DampingBase
import weakref 
from abc import ABC, abstractmethod
from typing import List, Dict, Type, Optional, Union

class RegionBase(ABC):
    _regions = {}  # Dictionary to hold regions
    _global_region = None  # Global region instance

    def __new__(cls, tag=None, damping: Type[DampingBase] = None):

        if tag in cls._regions:
            return cls._regions[tag]
        
        if tag == 0:
            cls._global_region = super().__new__(cls)
            cls._regions[0] = cls._global_region
            return cls._global_region

        # Create a new instance for a unique tag
        instance = super().__new__(cls)
        return instance



    def __init__(self, tag=None, damping: Type[DampingBase] = None):
        
        if tag == 0:
            # check the region is type of global region for tag 0
            if not isinstance(self, GlobalRegion):
                raise ValueError("Region with tag 0 should be of type GlobalRegion \
                                 please use GlobalRegion class to create a global region")
            self._tag = tag
            self._name = "GlobalRegion"
            self._regions[0] = self
            self.damping = damping
        elif tag is None:
            tag = self._get_next_tag()
            self._tag = tag
            self._name = f"region {tag}"
            self._regions[tag] = self
            self.damping = damping

        # Default values
        if not hasattr(self, "active"):
            self.active = True

    @property
    def tag(self):
        return self._tag
    
    @property
    def name(self):
        return self._name
    
    @property
    def damping(self):
        return self._damping() if self._damping else None

    @damping.setter
    def damping(self, value: Optional[Type[DampingBase]]):
        if value is not None and not issubclass(value.__class__, DampingBase):
            raise TypeError("damping must be a subclass of DampingBase")
        self._damping = weakref.ref(value) if value else None 

    @damping.deleter
    def damping(self):
        self._damping_ref = None 


    def __str__(self) -> str:
        type_name = self.get_type()
        res = f"Region class"
        res += f"\n\tTag: {self.tag}"
        res += f"\n\tName: {self.name}"
        res += f"\n\tType: {type_name}"
        res += f"\n\tActive: {self.active}"
        res += f"\n\tDamping: " + ("\n\t" + str(self.damping)).replace("\n", "\n\t") if self.damping else "\n\tDamping: None"
        return res

    @classmethod
    def _get_next_tag(cls):
        """Get the next available tag."""
        return len(cls._regions)

    @classmethod
    def _update_tags(cls):
        """Update tags and names of all regions except the global region."""
        tags = sorted(cls._regions.keys())
        if 0 in tags:
            tags.remove(0)  # Preserve the global region tag
        new_regions = {0: cls._regions[0]} if 0 in cls._regions else {}
        for i, old_tag in enumerate(tags, start=1):
            region = cls._regions[old_tag]
            region._tag = i
            region._name = f"region{i}"
            new_regions[i] = region
        cls._regions = new_regions

    @classmethod
    def remove_region(cls, tag):
        """Remove a region by its tag."""
        if tag == 0:
            raise ValueError("Cannot remove the global region (tag 0)")
        if tag in cls._regions:
            del cls._regions[tag]
            cls._update_tags()

    @classmethod
    def get_region(cls, tag) -> Optional["RegionBase"]:
        """Get a region by its tag."""
        return cls._regions.get(tag)

    @classmethod
    def get_all_regions(cls) -> Dict[int, "RegionBase"]:
        """Get all regions."""
        return cls._regions

    @classmethod
    def clear(cls) -> None:
        """Clear all regions, including the global region."""
        cls._regions = {}
        cls._global_region = None

    @staticmethod
    def print_regions() -> None:
        """Print all regions."""
        for tag, region in RegionBase.get_all_regions().items():
            print(region)


    @abstractmethod
    def to_tcl(self) -> str:
        """Convert the region to a TCL representation."""
        pass

    @abstractmethod
    def validate(self):
        """Validate region parameters."""
        pass

    @staticmethod
    @abstractmethod
    def get_Parameters() -> Dict[str, str]:
        """Get parameters for the region."""
        pass

    @staticmethod
    @abstractmethod
    def getNotes() -> Dict[str, list[str]]:
        """Get notes for the region."""
        pass

    @staticmethod
    @abstractmethod
    def get_type(self) -> str:
        pass

    @abstractmethod
    def setComponent(self, component:str, value:list[int]):
        pass


class GlobalRegion(RegionBase):
    def __new__(cls, damping: Type[DampingBase] = None):
        return super().__new__(cls, tag=0)

    def __init__(self, damping: Type[DampingBase] = None):
        super().__init__(tag=0, damping=damping)
        self.elements = None
        self.element_range = None
        self.nodes = None
        self.node_range = None

    def to_tcl(self) -> str:
        """TCL representation for global region."""
        return "region 0"

    def validate(self):
        """No validation required for global region."""
        pass

    @staticmethod
    def get_Parameters() -> Dict[str, str]:
        """No parameters for global region."""
        return {}

    @staticmethod
    def getNotes() -> Dict[str, list[str]]:
        """Notes about global region."""
        # return "Global region representing entire model"
        return {
            "Notes": [
                "Global region representing entire model"
            ],
            "References": []    
        }
    @staticmethod
    def get_type() -> str:
        return "GlobalRegion"
    
    def setComponent(self, component:str, values:list[int]):
        if component == "element":
            self.elements = values
            self.element_range = None
            self.nodes = None
            self.node_range = None
        elif component == "elementRange":
            if len(values) != 2:
                raise ValueError("element_range should have 2 elements")
            self.element_range = values
            self.elements = None
            self.nodes = None
            self.node_range = None
        elif component == "node":
            self.nodes = values
            self.node_range = None
            self.elements = None
            self.element_range = None
        elif component == "nodeRange":
            if len(values) != 2:
                raise ValueError("node_range should have 2 elements")
            self.node_range = values
            self.nodes = None
            self.elements = None
            self.element_range = None
        else:
            raise ValueError(f"""Invalid component {component} for GlobalRegion
                             valid components are element, elementRange, node, nodeRange""")
        

def initialize_region_base():
    """Initialize the RegionBase with the global region."""
    GlobalRegion()


class ElementRegion(RegionBase):
    def __new__(cls, damping: Type[DampingBase] = None, **kwargs):
        return super().__new__(cls, tag=kwargs.pop("tag", None))
    
    def __init__(self, damping: Type[DampingBase] = None ,**kwargs):
        super().__init__(tag=kwargs.pop("tag", None), damping=damping)
        self.elements = []
        self.element_range = []
        self.element_only = False

        if kwargs:
            validated = self.validate(**kwargs)
            self.elements = validated["elements"]
            self.element_range = validated["element_range"]
            self.element_only = validated["element_only"]


    def to_tcl(self):
        cmd = f"eval \"region {self.tag}"
        if len(self.element_range) > 0:
            cmd += " -eleRange {} {}".format(*self.element_range)
            if self.element_only:
                cmd += "Only"
        elif len(self.elements) > 0:
            cmd += " -ele" + ("Only" if self.element_only else "")
            cmd += " " + " ".join(str(e) for e in self.elements)

        if self.damping:
            if self.damping.get_Type() in ["RayleighDamping", "Frequency Rayleigh"]:
                cmd += f" -rayleigh {self.damping.alphaM} {self.damping.betaK} {self.damping.betaKInit} {self.damping.betaKComm}"
            else:
                cmd += f" -damping {self.damping.tag}"
        cmd += "\""
        return cmd
    

    def __str__(self):
        res = super().__str__()
        res += f"\n\tNum of Elements: {len(self.elements)}"
        res += f"\n\tElement Range: {self.element_range}"
        res += f"\n\tElement Only: {self.element_only}"
        return res


    @staticmethod
    def get_Parameters():
        return {
            "elements": "lineEdit",
            "element_range": "lineEdit",
            "element_only": "checkbox"
        }
    
    @staticmethod
    def getNotes()->Dict[str, list[str]]:
        return {
            "Notes": [
                "Use elements list for specific elements: [1, 2, 3, ...]"
                "If you use GUI, use comma separated values 1, 2, 3, ...",
                "Use element_range for a range: [start, end]",
                "If you use GUI, use comma separated values start, end for range",
                "Cannot use both elements and element_range simultaneously",
                "Set element_only=True to include only elements (default is False)",
                "If you use GUI, check the checkbox to include only elements"
            ],
            "References": []
        }

    @staticmethod
    def validate(**kwargs) -> Dict[str, Union[str, bool, list]]:
        # check if elements or element_range is provided
        # if "elements" not in kwargs and "element_range" not in kwargs:
        #     raise ValueError("Either elements or element_range should be provided")
        
        # check if elements and element_range are provided at the same time
        if "elements" in kwargs and "element_range" in kwargs:
            raise ValueError("Both elements and element_range cannot be provided at the same time")
        
        elements = []
        element_range = []
        element_only = False

        if "elements" in kwargs:
            elements = kwargs["elements"]
            # elements should be list of integers check that
            if not isinstance(elements, list):
                raise ValueError("elements should be a list of integers")
            for e in elements:
                if not isinstance(e, int):
                    raise ValueError("elements should be a list of integers")
                

        if "element_range" in kwargs:
            element_range = kwargs["element_range"]
            if not isinstance(element_range, list):
                raise ValueError("element_range should be a list of integers")
            if len(element_range) != 2:
                raise ValueError("element_range should have 2 elements")
            for e in element_range:
                if not isinstance(e, int):
                    raise ValueError("element_range should be a list of integers")

        if "element_only" in kwargs:
            element_only = kwargs["element_only"]
            if not isinstance(element_only, bool):
                raise ValueError("element_only should be a boolean")
            
        return {
            "elements": elements,
            "element_range": element_range,
            "element_only": element_only
        }
    
    @staticmethod
    def get_type() -> str:
        return "ElementRegion"
    
    def setComponent(self, component:str, value:list[int]):
        if component == "element":
            self.elements = value
            self.element_range = []
        elif component == "elementRange":
            if len(value) != 2:
                raise ValueError("element_range should have 2 elements")
            self.element_range = value
            self.elements = []
        else:
            raise ValueError(f"""Invalid component {component} for ElementRegion 
                             valid components are element and elementRange""")


class NodeRegion(RegionBase):
    def __new__(cls, damping: Type[DampingBase] = None, **kwargs):
        return super().__new__(cls, tag=kwargs.pop("tag", None))
    
    def __init__(self, damping: Type[DampingBase] = None, **kwargs):
        super().__init__(tag=kwargs.pop("tag", None), damping=damping)
        self.nodes = []
        self.node_range = []
        self.node_only = False
        
        if kwargs:
            validated = self.validate(**kwargs)
            self.nodes = validated["nodes"]
            self.node_range = validated["node_range"]
            self.node_only = validated["node_only"]


    def to_tcl(self):
        cmd = f"region {self.tag}"
        if len(self.node_range) > 0:
            cmd += " -node" + ("Only" if self.node_only else "") + "Range"
            cmd += " {} {}".format(*self.node_range)
        elif len(self.nodes) > 0:
            cmd += " -node" + ("Only" if self.node_only else "")
            cmd += " " + " ".join(str(n) for n in self.nodes)

        if self.damping:
            if self.damping.get_Type() in ["RayleighDamping", "Frequency Rayleigh"]:
                cmd += F"-rayleigh {self.damping.alphaM} {self.damping.betaK} {self.damping.betaKInit} {self.damping.betaKComm}"
            else:
                cmd += f"-damping {self.damping.tag}"
        return cmd
    


    def __str__(self):
        res = super().__str__()
        res += f"\n\tNum of Nodes: {len(self.nodes)}"
        res += f"\n\tNode Range: {self.node_range}"
        return res

    @staticmethod
    def get_Parameters():
        return {
            "nodes": "lineEdit",
            "node_range": "lineEdit"
        }

    @staticmethod
    def getNotes()->Dict[str, list[str]]:
        return {
            "Notes": [
                "Use nodes list for specific nodes: [1, 2, 3, ...]",
                "If you use GUI, use comma separated values 1, 2, 3, ...",
                "Use node_range for a range: [start, end]",
                "If you use GUI, use comma separated values start, end for range",
                "Cannot use both nodes and node_range simultaneously",
                "Set node_only=True to include only nodes (default is False)",
                "If you use GUI, check the checkbox to include only nodes",
            ],
            "References": []
        }

    @staticmethod
    def validate(**kwargs) -> Dict[str, Union[str, list]]:
        # if "nodes" not in kwargs and "node_range" not in kwargs:
        #     raise ValueError("Either nodes or node_range should be provided")
        
        if "nodes" in kwargs and "node_range" in kwargs:
            raise ValueError("Both nodes and node_range cannot be provided at the same time")
        
        nodes = []
        node_range = []
        node_only = False

        if "nodes" in kwargs:
            nodes = kwargs["nodes"]
            if not isinstance(nodes, list):
                raise ValueError("nodes should be a list of integers")
            for n in nodes:
                if not isinstance(n, int):
                    raise ValueError("nodes should be a list of integers")

        if "node_range" in kwargs:
            node_range = kwargs["node_range"]
            if not isinstance(node_range, list):
                raise ValueError("node_range should be a list of integers")
            if len(node_range) != 2:
                raise ValueError("node_range should have 2 elements")
            for n in node_range:
                if not isinstance(n, int):
                    raise ValueError("node_range should be a list of integers")
                
        if "node_only" in kwargs:
            node_only = kwargs["node_only"]
            if not isinstance(node_only, bool):
                raise ValueError("node_only should be a boolean")
            
        return {
            "nodes": nodes,
            "node_range": node_range,
            "node_only": node_only
        }

    @staticmethod
    def get_type() -> str:
        return "NodeRegion"
    
    def setComponent(self, component:str, value:list[int]):
        if component == "node":
            self.nodes = value
            self.node_range = []
        elif component == "nodeRange":
            if len(value) != 2:
                raise ValueError("node_range should have 2 elements")
            self.node_range = value
            self.nodes = []
        else:
            raise ValueError(f"""Invalid component {component} for NodeRegion
                             valid components are node and nodeRange""")

class RegionManager:
    """
    A centralized manager for all region instances in MeshMaker.
    
    The RegionManager implements the Singleton pattern to ensure a single, consistent
    point of region management across the entire application. It provides methods for
    creating, retrieving, and managing region objects used to define specific parts
    of structural models for analysis and damping assignments.
    
    All region objects created through this manager are automatically tracked and tagged,
    simplifying the process of defining and managing model regions. A special GlobalRegion
    with tag 0 is automatically created when the RegionManager is initialized.
    
    """
    _instance = None
    def __new__(cls):
        """
        Create a new RegionManager instance or return the existing one if already created.
        
        Returns:
            RegionManager: The singleton instance of the RegionManager
        """
        if cls._instance is None:
            cls._instance = super(RegionManager, cls).__new__(cls)
            cls._instance._initialized = False
        return cls._instance

    def __init__(self):
        """
        Initialize the RegionManager and create the GlobalRegion.
        
        This method ensures that the global region (tag 0) is automatically created
        when the RegionManager is initialized. The initialization only happens once
        due to the singleton pattern.
        """
        if not self._initialized:
            self._initialized = True
            initialize_region_base()

    @property
    def regions(self):
        """
        Get all regions as a dictionary mapping tags to region instances.
        
        Returns:
            Dict[int, RegionBase]: A dictionary mapping tags to region instances
        """
        return RegionBase.get_all_regions()

    def get_region(self, tag):
        """
        Get a region by its tag.
        
        Args:
            tag (int): The unique identifier of the region instance
            
        Returns:
            RegionBase: The region instance with the specified tag, or None if not found
        """
        return RegionBase.get_region(tag)
    
    def create_region(self, regionType: str, damping=None, **kwargs):
        """
        Create a new region instance of the specified type.
        
        This method creates and returns a new region object based on the provided type
        and parameters. The region is automatically registered and tracked.
        
        Args:
            regionType (str): The type of region to create. Available types include:
                'ElementRegion': A region defined by a set of elements or element range
                'NodeRegion': A region defined by a set of nodes or node range
                'GlobalRegion': The global region representing the entire model
            damping: Optional damping instance to associate with the region
            **kwargs: Specific parameters for the region type being created
                (e.g., elements, element_range for ElementRegion,
                nodes, node_range for NodeRegion)
        
        Returns:
            RegionBase: A new instance of the requested region type
            
        Raises:
            ValueError: If the region type is unknown or if required parameters are missing or invalid
        """
        if regionType.lower() == "elementregion":
            return ElementRegion(damping=damping, **kwargs)
        elif regionType.lower() == "noderegion":
            return NodeRegion(damping=damping, **kwargs)
        elif regionType.lower() == "globalregion":
            return GlobalRegion(damping=damping)
        else:
            raise ValueError(f"""Invalid region type {regionType}
                             valid region types are ElementRegion, NodeRegion, GlobalRegion""")

    def remove_region(self, tag):
        """
        Remove a region by its tag.
        
        This method removes the specified region instance and automatically
        updates the tags of the remaining region instances to maintain sequential numbering.
        The GlobalRegion (tag 0) cannot be removed.
        
        Args:
            tag (int): The unique identifier of the region instance to remove
            
        Raises:
            ValueError: If attempting to remove the GlobalRegion (tag 0)
        """
        RegionBase.remove_region(tag)

    def clear_regions(self):
        """
        Remove all regions and reinitialize the GlobalRegion.
        
        This method clears all region instances, including the GlobalRegion,
        and then recreates the GlobalRegion with tag 0.
        """
        RegionBase.clear()
        initialize_region_base()

    def print_regions(self):
        """
        Print information about all regions.
        
        This method prints detailed information about all region instances
        including their tags, types, elements/nodes, and damping assignments.
        """
        RegionBase.print_regions()

if __name__ == "__main__":
    from femora.components.Damping.dampingBase import RayleighDamping, ModalDamping
    # ---- Test Global Region ----
    # test the DampingBase class
    damping1 = DampingBase()
    damping2 = DampingBase()
    damping3 = DampingBase()

    RayleighDamping1 = RayleighDamping(alphaM=0.1, betaK=0.2, betaKInit=0.3, betaKComm=0.4)
    RayleighDamping2 = RayleighDamping(alphaM=0.5, betaK=0.6, betaKInit=0.7, betaKComm=0.8)
    ModalDamping1 = ModalDamping(numberofModes=2, dampingFactors="0.1,0.2")
    ModalDamping2 = ModalDamping(numberofModes=3, dampingFactors="0.3,0.4,0.5")

    initialize_region_base()
    global_region = RegionBase.get_region(0)
    global_region2 = GlobalRegion()
    global_region3 = GlobalRegion()



    assert global_region == global_region2, "Global regions should be equal"
    assert global_region is global_region3, "Global regions should be the same instance"
    assert global_region is not None, "Global region should exist"
    assert global_region.tag == 0, "Global region tag should be 0"
    assert global_region.name == "GlobalRegion", "Global region name incorrect"
    global_region.damping = DampingBase.get_damping(4)


    eleregion1 = ElementRegion(elements=[1, 2, 3])
    print("Global Region tests passed")



    eleregion1 = ElementRegion(elements=[1, 2, 3])
    eleregion2 = ElementRegion(element_range=[4, 10])
    eleregion4 = ElementRegion()
    eleregion5 = ElementRegion(tag=0)

    # Test Element Regions
    assert eleregion1.tag == 2, "Element region 1 tag should be 1"
    assert eleregion2.tag == 3, "Element region 2 tag should be 2"
    assert len(eleregion1.elements) == 3, "Element region 1 should have 3 elements"
    assert eleregion2.element_range == [4, 10], "Element region 2 range incorrect"
    eleregion3 = ElementRegion(tag=2)
    assert eleregion3.tag == 2, "Element region 3 tag should be 2"
    assert len(eleregion3.elements) == 0, "Element region 3 should have 0 elements"
    print("Element Region tests passed")

    # Test Node Regions
    noderegion1 = NodeRegion(nodes=[1, 2, 3])
    noderegion2 = NodeRegion(node_range=[4, 10])

    assert noderegion1.tag == 5, "Node region 1 tag should be 3"
    assert noderegion2.tag == 6, "Node region 2 tag should be 4"
    assert len(noderegion1.nodes) == 3, "Node region 1 should have 3 nodes"
    assert noderegion2.node_range == [4, 10], "Node region 2 range incorrect"
    noderegion3 = NodeRegion(tag=5)
    assert noderegion3.tag == 5, "Node region 3 tag should be 3"
    print("Node Region tests passed")

    # Test region removal and tag updates
    # Test removing global region (should raise ValueError)
    try:
        RegionBase.remove_region(0)
        assert False, "Should have raised ValueError"
    except ValueError:
        assert True

    # assign with damping 
    eleDamping = ElementRegion(elements=[1, 2, 4], damping=RayleighDamping1)
    nodeDamping = NodeRegion(nodes=[1, 2, 4], damping=ModalDamping1)
    baseDamping = GlobalRegion(damping=RayleighDamping1)
    # print(eleDamping)
    # print(nodeDamping)
    # print(baseDamping)

    print("Region assignment with damping passed")
    
    # Test damping assignment
    noderegion1.damping = RayleighDamping2
    assert noderegion1.damping is not None, "Damping should be assigned"
    print("Damping assignment test passed")

    print("\nAll tests passed successfully!")





from qtpy.QtWidgets import (
    QWidget, QVBoxLayout, QTabWidget, QApplication, QScrollArea
)
from qtpy.QtCore import Qt

from femora.components.DRM.drmAbsorbingGUI import AbsorbingGUI
from femora.components.DRM.drmPatternGUI import DRMGUI
from femora.components.DRM.drmProcessGUI import DRMProcessGUI

class CombinedDRMGUI(QWidget):
    """
    A unified GUI that combines all DRM-related interfaces:
    - Pattern loading interface for H5DRM files
    - Absorbing layer management
    - DRM process configuration
    """
    def __init__(self, parent=None):
        """
        Initialize the combined DRM GUI.
        
        Args:
            parent: Parent widget
        """
        super().__init__(parent)
        
        # Create main layout
        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        
        # Create tab widget to organize different DRM components
        self.tabWidget = QTabWidget()
        
        # Create individual DRM component widgets
        self.drmGUI = DRMGUI()
        self.absorbingGUI = AbsorbingGUI()
        self.processGUI = DRMProcessGUI()
        
        # Create scroll areas for each component to ensure they can be viewed
        # even in limited screen space
        
        # DRM Pattern tab
        drmPatternScroll = QScrollArea()
        drmPatternScroll.setWidgetResizable(True)
        drmPatternScroll.setWidget(self.drmGUI)
        
        # Absorbing Layer tab
        absorbingScroll = QScrollArea()
        absorbingScroll.setWidgetResizable(True)
        absorbingScroll.setWidget(self.absorbingGUI)
        
        # Process tab
        processScroll = QScrollArea()
        processScroll.setWidgetResizable(True)
        processScroll.setWidget(self.processGUI)
        
        # Add tabs with scroll areas
        self.tabWidget.addTab(drmPatternScroll, "DRM Pattern")
        self.tabWidget.addTab(absorbingScroll, "Absorbing Layer")
        self.tabWidget.addTab(processScroll, "DRM Process")
        
        # Add tab widget to the main layout
        layout.addWidget(self.tabWidget)
    
    def set_active_tab(self, tab_name):
        """
        Set the active tab based on name
        
        Args:
            tab_name (str): Name of tab to activate ("DRM Pattern", "Absorbing Layer", or "DRM Process")
        """
        tab_map = {
            "DRM Pattern": 0,
            "Absorbing Layer": 1,
            "DRM Process": 2
        }
        
        if tab_name in tab_map:
            self.tabWidget.setCurrentIndex(tab_map[tab_name])
    
    # Convenience methods to directly access underlying components
    
    def get_drm_gui(self):
        """Get the DRM pattern GUI component"""
        return self.drmGUI
    
    def get_absorbing_gui(self):
        """Get the absorbing layer GUI component"""
        return self.absorbingGUI
    
    def get_process_gui(self):
        """Get the DRM process GUI component"""
        return self.processGUI


if __name__ == '__main__':
    import sys
    app = QApplication(sys.argv)
    window = CombinedDRMGUI()
    window.show()
    sys.exit(app.exec_())



from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QWidget, QVBoxLayout, QHBoxLayout, QLabel, QLineEdit, 
    QComboBox, QPushButton, QTableWidget, QTableWidgetItem, 
    QDialog, QFormLayout, QMessageBox, QHeaderView, QGridLayout, QMenu,
    QSizePolicy, QProgressDialog, QApplication, QSpinBox, QDoubleSpinBox
)
from femora.components.Constraint.mpConstraint import (
    mpConstraint, equalDOF, rigidLink, rigidDiaphragm, mpConstraintManager
)
import time

class MPConstraintManagerTab(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        
        # Configure dialog size and properties
        self.setWindowTitle("MP Constraints Manager")
        self.resize(800, 600)
        
        # Main layout
        layout = QVBoxLayout(self)
        layout.setSpacing(10)
        layout.setContentsMargins(10, 10, 10, 10)
        
        # Constraint type selection
        type_layout = QGridLayout()
        
        # Constraint type dropdown
        self.constraint_type_combo = QComboBox()
        self.constraint_type_combo.addItems(["equalDOF", "rigidLink", "rigidDiaphragm", "laminarBoundary"])
        
        create_constraint_btn = QPushButton("Create New Constraint")
        create_constraint_btn.clicked.connect(self.open_constraint_creation_dialog)
        
        type_layout.addWidget(QLabel("Constraint Type:"), 0, 0)
        type_layout.addWidget(self.constraint_type_combo, 0, 1)
        type_layout.addWidget(create_constraint_btn, 1, 0, 1, 2)
        
        layout.addLayout(type_layout)
        
        # Status information showing total constraints
        self.status_label = QLabel("Loading constraints...")
        layout.addWidget(self.status_label)
        
        # Constraints table
        self.constraints_table = QTableWidget()
        self.constraints_table.setColumnCount(5)  # Tag, Type, Master, Slaves, Parameters
        self.constraints_table.setHorizontalHeaderLabels(["Tag", "Type", "Master", "Slaves", "Parameters"])
        
        # Configure table appearance
        self.constraints_table.verticalHeader().setVisible(False)  # Hide row numbers
        self.constraints_table.setMinimumWidth(750)
        self.constraints_table.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        
        # Set horizontal header properties
        header = self.constraints_table.horizontalHeader()
        header.setStretchLastSection(True)
        header.setSectionResizeMode(QHeaderView.Interactive)
        
        # Set default column widths
        self.constraints_table.setColumnWidth(0, 60)   # Tag column
        self.constraints_table.setColumnWidth(1, 120)  # Type column
        self.constraints_table.setColumnWidth(2, 100)  # Master column
        self.constraints_table.setColumnWidth(3, 200)  # Slaves column
        
        # Configure double-click and context menu
        self.constraints_table.setContextMenuPolicy(Qt.CustomContextMenu)
        self.constraints_table.customContextMenuRequested.connect(self.show_context_menu)
        self.constraints_table.cellDoubleClicked.connect(self.handle_double_click)
        
        layout.addWidget(self.constraints_table)
        
        # Pagination controls
        pagination_layout = QHBoxLayout()
        
        # Page size selection
        pagination_layout.addWidget(QLabel("Items per page:"))
        self.page_size_combo = QComboBox()
        self.page_size_combo.addItems(["100", "500", "1000", "All"])
        self.page_size_combo.setCurrentText("100")
        self.page_size_combo.currentTextChanged.connect(self.change_page_size)
        pagination_layout.addWidget(self.page_size_combo)
        
        # Page navigation with spinbox
        pagination_layout.addStretch()
        pagination_layout.addWidget(QLabel("Go to page:"))
        
        self.page_spinbox = QSpinBox()
        self.page_spinbox.setMinimum(1)
        self.page_spinbox.setMaximum(1)  # Will be updated with total pages
        self.page_spinbox.setValue(1)
        self.page_spinbox.valueChanged.connect(self.go_to_page)
        pagination_layout.addWidget(self.page_spinbox)
        
        self.page_label = QLabel("of 1")
        pagination_layout.addWidget(self.page_label)
        
        layout.addLayout(pagination_layout)
        
        button_layout = QVBoxLayout()
        # Refresh button
        refresh_btn = QPushButton("Refresh Constraints List")
        refresh_btn.clicked.connect(self.refresh_constraints_list)
        
        # Add Clear All button
        button_layout.addWidget(refresh_btn)
        # Clear All button with warning color styling
        clear_all_btn = QPushButton("Clear All Constraints")
        clear_all_btn.clicked.connect(self.clear_all_constraints)
        button_layout.addWidget(clear_all_btn)
        
        layout.addLayout(button_layout)
        
        # Pagination state
        self.current_page = 1
        self.total_pages = 1
        self.page_size = 100
        
        # Initial refresh
        self.refresh_constraints_list()

    def change_page_size(self, size_text):
        """Change the number of items loaded per page"""
        if size_text == "All":
            self.page_size = len(mpConstraint._constraints)
        else:
            try:
                self.page_size = int(size_text)
            except ValueError:
                self.page_size = 100
        
        # Reset to first page and refresh
        self.current_page = 1
        self.refresh_constraints_list()

    def open_constraint_creation_dialog(self):
        """Open dialog to create new constraint"""
        constraint_type = self.constraint_type_combo.currentText()
        
        if constraint_type == "laminarBoundary":
            dialog = MPLaminarBoundaryDialog(self)
        else:
            dialog = MPConstraintCreationDialog(constraint_type, self)
        
        if dialog.exec() == QDialog.Accepted:
            self.refresh_constraints_list()

    def refresh_constraints_list(self):
        """Update constraints table with current data for the current page"""
        # Show progress dialog for large constraint sets
        count = len(mpConstraint._constraints)
        
        if count > 1000:
            progress = QProgressDialog("Loading constraints...", "Cancel", 0, 100, self)
            progress.setWindowModality(Qt.WindowModal)
            progress.setValue(0)
            progress.show()
            QApplication.processEvents()
        
        # Calculate total pages
        if self.page_size > 0:
            self.total_pages = max(1, (count + self.page_size - 1) // self.page_size)
        else:
            self.total_pages = 1
        
        # Ensure current page is valid
        if self.current_page > self.total_pages:
            self.current_page = self.total_pages
        
        # Update page label
        self.page_label.setText(f"of {self.total_pages}")
        
        # Update page spinbox without triggering the valueChanged signal
        self.page_spinbox.blockSignals(True)
        self.page_spinbox.setMaximum(self.total_pages)
        self.page_spinbox.setValue(self.current_page)
        self.page_spinbox.blockSignals(False)
        
        # Clear the table
        self.constraints_table.setRowCount(0)
        
        # Get all constraints as a list for pagination
        all_constraints = list(mpConstraint._constraints.values())
        
        # Calculate start and end indices for the current page
        start_idx = (self.current_page - 1) * self.page_size
        end_idx = min(start_idx + self.page_size, len(all_constraints))
        
        # Get constraints for the current page
        page_constraints = all_constraints[start_idx:end_idx]
        
        # Set row count for the current page
        self.constraints_table.setRowCount(len(page_constraints))
        
        # Update progress to 20%
        if count > 1000:
            progress.setValue(20)
            QApplication.processEvents()
        
        # Process constraints in batches of 100 to maintain UI responsiveness
        batch_size = 100
        for batch_start in range(0, len(page_constraints), batch_size):
            batch_end = min(batch_start + batch_size, len(page_constraints))
            batch = page_constraints[batch_start:batch_end]
            
            for row, constraint in enumerate(batch, batch_start):
                # Tag
                tag_item = QTableWidgetItem(str(constraint.tag))
                tag_item.setFlags(tag_item.flags() & ~Qt.ItemIsEditable)
                tag_item.setData(Qt.UserRole, constraint.tag)  # Store tag for reference
                self.constraints_table.setItem(row, 0, tag_item)
                
                # Type
                type_item = QTableWidgetItem(constraint.__class__.__name__)
                type_item.setFlags(type_item.flags() & ~Qt.ItemIsEditable)
                self.constraints_table.setItem(row, 1, type_item)
                
                # Master Node
                master_item = QTableWidgetItem(str(constraint.master_node))
                master_item.setFlags(master_item.flags() & ~Qt.ItemIsEditable)
                self.constraints_table.setItem(row, 2, master_item)
                
                # Slave Nodes
                slaves_str = ", ".join(map(str, constraint.slave_nodes))
                slaves_item = QTableWidgetItem(slaves_str)
                slaves_item.setFlags(slaves_item.flags() & ~Qt.ItemIsEditable)
                self.constraints_table.setItem(row, 3, slaves_item)
                
                # Parameters
                params = ""
                if isinstance(constraint, equalDOF):
                    params = f"DOFs: {', '.join(map(str, constraint.dofs))}"
                elif isinstance(constraint, rigidLink):
                    params = f"Type: {constraint.type}"
                elif isinstance(constraint, rigidDiaphragm):
                    params = f"Direction: {constraint.direction}"
                params_item = QTableWidgetItem(params)
                params_item.setFlags(params_item.flags() & ~Qt.ItemIsEditable)
                self.constraints_table.setItem(row, 4, params_item)
            
            # Update progress for each batch
            if count > 1000:
                progress_value = 20 + int(80 * (batch_end / len(page_constraints)))
                progress.setValue(progress_value)
                QApplication.processEvents()
                
                if progress.wasCanceled():
                    progress.close()
                    return  # Exit the function early to avoid partial updates
        
        # Update status label with explicit page information
        self.status_label.setText(f"Total Constraints: {count} | Showing {len(page_constraints)} constraints | Currently on Page {self.current_page} of {self.total_pages}")
        
        # Close progress dialog if it was shown
        if count > 1000:
            progress.setValue(100)
            progress.close()

    def go_to_page(self, page):
        """Go to the specified page number"""
        if 1 <= page <= self.total_pages:
            self.current_page = page
            self.refresh_constraints_list()

    def handle_double_click(self, row, column):
        """Handle double-click on a table cell to edit constraint"""
        tag_item = self.constraints_table.item(row, 0)
        if tag_item:
            tag = tag_item.data(Qt.UserRole)
            constraint = mpConstraint._constraints.get(tag)
            if constraint:
                self.open_constraint_edit_dialog(constraint)

    def show_context_menu(self, position):
        """Show context menu for right-clicked table cell"""
        # Get the row under the mouse
        row = self.constraints_table.rowAt(position.y())
        if row < 0:
            return
        
        # Get the constraint tag
        tag_item = self.constraints_table.item(row, 0)
        if not tag_item:
            return
        
        tag = tag_item.data(Qt.UserRole)
        
        # Create context menu
        menu = QMenu(self)
        edit_action = menu.addAction("Edit Constraint")
        delete_action = menu.addAction("Delete Constraint")
        
        # Connect actions
        action = menu.exec_(self.constraints_table.mapToGlobal(position))
        if action == edit_action:
            constraint = mpConstraint._constraints.get(tag)
            if constraint:
                self.open_constraint_edit_dialog(constraint)
        elif action == delete_action:
            self.delete_constraint(tag)

    def open_constraint_edit_dialog(self, constraint):
        """Open dialog to edit existing constraint"""
        dialog = MPConstraintEditDialog(constraint, self)
        if dialog.exec() == QDialog.Accepted:
            self.refresh_constraints_list()

    def delete_constraint(self, tag):
        """Delete constraint from system"""
        reply = QMessageBox.question(
            self, 'Delete Constraint',
            f"Delete constraint with tag {tag}?",
            QMessageBox.Yes | QMessageBox.No
        )
        if reply == QMessageBox.Yes:
            mpConstraintManager().remove_constraint(tag)
            self.refresh_constraints_list()
            
    def clear_all_constraints(self):
        """Delete all constraints from the system"""
        count = len(mpConstraint._constraints)
        
        if count == 0:
            QMessageBox.information(self, "No Constraints", "There are no constraints to clear.")
            return
            
        reply = QMessageBox.question(
            self, 'Clear All Constraints',
            f"Delete all {count} constraints? This action cannot be undone.",
            QMessageBox.Yes | QMessageBox.No
        )
        
        if reply == QMessageBox.Yes:
            # Show progress dialog for large numbers of constraints
            if count > 100:
                progress = QProgressDialog("Deleting all constraints...", "Cancel", 0, 100, self)
                progress.setWindowModality(Qt.WindowModal)
                progress.setValue(0)
                progress.show()
                QApplication.processEvents()
            
            # Clear the constraints dictionary directly
            mpConstraint._constraints.clear()
            
            # Close progress dialog if it was shown
            if count > 100:
                progress.setValue(100)
                progress.close()
                
            self.refresh_constraints_list()
            QMessageBox.information(self, "Success", f"Successfully cleared {count} constraints.")


# New Laminar Boundary Dialog
class MPLaminarBoundaryDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Create Laminar Boundary Constraint")
        self.manager = mpConstraintManager()
        
        layout = QVBoxLayout(self)
        form_layout = QFormLayout()
        
        # DOFs (degrees of freedom)
        self.dofs_input = QLineEdit("1 2")
        self.dofs_input.setPlaceholderText("Enter DOFs (e.g., 1 2 3)")
        form_layout.addRow("DOFs (space separated):", self.dofs_input)
        
        # Direction
        self.direction_combo = QComboBox()
        self.direction_combo.addItems(["1 (X)", "2 (Y)", "3 (Z)"])
        self.direction_combo.setCurrentIndex(2)  # Default to Z direction (3)
        form_layout.addRow("Perpendicular Direction:", self.direction_combo)
        
        # Tolerance
        self.tolerance_spin = QDoubleSpinBox()
        self.tolerance_spin.setRange(1e-6, 1.0)
        self.tolerance_spin.setValue(0.01)
        self.tolerance_spin.setSingleStep(0.001)
        self.tolerance_spin.setDecimals(4)
        form_layout.addRow("Tolerance:", self.tolerance_spin)
        
        # Description
        description = QLabel(
            "Laminar boundary will connect nodes on the boundary of the mesh that have\n"
            "the same coordinate in the specified direction within the given tolerance.\n"
            "This creates Equal DOF constraints along planar boundaries."
        )
        description.setWordWrap(True)
        form_layout.addRow(description)
        
        layout.addLayout(form_layout)
        
        # Buttons
        btn_layout = QHBoxLayout()
        create_btn = QPushButton("Create")
        create_btn.clicked.connect(self.create_laminar_boundary)
        cancel_btn = QPushButton("Cancel")
        cancel_btn.clicked.connect(self.reject)
        
        btn_layout.addWidget(create_btn)
        btn_layout.addWidget(cancel_btn)
        layout.addLayout(btn_layout)
    
    def create_laminar_boundary(self):
        try:
            # Get direction (1, 2, or 3)
            direction_text = self.direction_combo.currentText()
            direction = int(direction_text[0])  # Extract the first character and convert to int
            
            # Get DOFs
            dofs = list(map(int, self.dofs_input.text().split()))
            
            # Get tolerance
            tolerance = self.tolerance_spin.value()
            
            # Show progress dialog since this operation can be time-consuming
            progress = QProgressDialog("Creating laminar boundary constraints...", "Cancel", 0, 100, self)
            progress.setWindowModality(Qt.WindowModal)
            progress.show()
            QApplication.processEvents()
            
            # Create laminar boundary constraints
            try:
                # Create the constraints
                self.manager.create_laminar_boundary(dofs=dofs, direction=direction, tol=tolerance)
                progress.setValue(100)
                QMessageBox.information(self, "Success", "Laminar boundary constraints created successfully.")
                self.accept()
            except Exception as e:
                progress.close()
                QMessageBox.critical(self, "Error", f"Error creating laminar boundary: {str(e)}")
                return
        
        except ValueError as e:
            QMessageBox.warning(self, "Input Error", f"Invalid input: {str(e)}")


# Keep the original dialog classes without changes
class MPConstraintCreationDialog(QDialog):
    def __init__(self, constraint_type, parent=None):
        super().__init__(parent)
        self.setWindowTitle(f"Create {constraint_type} Constraint")
        self.constraint_type = constraint_type
        self.manager = mpConstraintManager()
        
        layout = QVBoxLayout(self)
        form_layout = QFormLayout()
        
        # Common fields
        self.master_input = QLineEdit()
        self.slaves_input = QLineEdit()
        
        form_layout.addRow("Master Node:", self.master_input)
        form_layout.addRow("Slave Nodes (comma separated):", self.slaves_input)
        
        # Type-specific fields
        self.type_specific_widget = QWidget()
        type_layout = QFormLayout(self.type_specific_widget)
        
        if constraint_type == "equalDOF":
            self.dofs_input = QLineEdit()
            type_layout.addRow("DOFs (space separated):", self.dofs_input)
        elif constraint_type == "rigidLink":
            self.type_combo = QComboBox()
            self.type_combo.addItems(["bar", "beam"])
            type_layout.addRow("Link Type:", self.type_combo)
        elif constraint_type == "rigidDiaphragm":
            self.direction_input = QLineEdit()
            type_layout.addRow("Direction:", self.direction_input)
        
        form_layout.addRow(self.type_specific_widget)
        layout.addLayout(form_layout)
        
        # Buttons
        btn_layout = QHBoxLayout()
        create_btn = QPushButton("Create")
        create_btn.clicked.connect(self.create_constraint)
        cancel_btn = QPushButton("Cancel")
        cancel_btn.clicked.connect(self.reject)
        
        btn_layout.addWidget(create_btn)
        btn_layout.addWidget(cancel_btn)
        layout.addLayout(btn_layout)

    def create_constraint(self):
        try:
            master = int(self.master_input.text())
            slaves = [int(n.strip()) for n in self.slaves_input.text().split(",")]
            
            args = [master, slaves]
            
            if self.constraint_type == "equalDOF":
                dofs = list(map(int, self.dofs_input.text().split()))
                args.append(dofs)
            elif self.constraint_type == "rigidLink":
                args.insert(0, self.type_combo.currentText())
            elif self.constraint_type == "rigidDiaphragm":
                args.insert(0, int(self.direction_input.text()))
            
            self.manager.create_constraint(self.constraint_type, *args)
            self.accept()
            
        except ValueError as e:
            QMessageBox.warning(self, "Input Error", f"Invalid input: {str(e)}")


class MPConstraintEditDialog(QDialog):
    def __init__(self, constraint, parent=None):
        super().__init__(parent)
        self.constraint = constraint
        self.manager = mpConstraintManager()
        self.setWindowTitle(f"Edit {type(constraint).__name__} Constraint")
        
        layout = QVBoxLayout(self)
        form_layout = QFormLayout()
        
        # Master Node
        self.master_input = QLineEdit(str(constraint.master_node))
        form_layout.addRow("Master Node:", self.master_input)
        
        # Slave Nodes
        self.slaves_input = QLineEdit(", ".join(map(str, constraint.slave_nodes)))
        form_layout.addRow("Slave Nodes:", self.slaves_input)
        
        # Type-specific fields
        if isinstance(constraint, equalDOF):
            self.dofs_input = QLineEdit(" ".join(map(str, constraint.dofs)))
            form_layout.addRow("DOFs:", self.dofs_input)
        elif isinstance(constraint, rigidLink):
            self.type_combo = QComboBox()
            self.type_combo.addItems(["bar", "beam"])
            self.type_combo.setCurrentText(constraint.type)
            form_layout.addRow("Type:", self.type_combo)
        elif isinstance(constraint, rigidDiaphragm):
            self.direction_input = QLineEdit(str(constraint.direction))
            form_layout.addRow("Direction:", self.direction_input)
        
        # Buttons
        btn_layout = QHBoxLayout()
        save_btn = QPushButton("Save")
        save_btn.clicked.connect(self.save_changes)
        cancel_btn = QPushButton("Cancel")
        cancel_btn.clicked.connect(self.reject)
        
        btn_layout.addWidget(save_btn)
        btn_layout.addWidget(cancel_btn)
        layout.addLayout(form_layout)
        layout.addLayout(btn_layout)

    def save_changes(self):
        try:
            # Update common properties
            new_master = int(self.master_input.text())
            new_slaves = [int(n.strip()) for n in self.slaves_input.text().split(",")]
            
            # Recreate constraint with new parameters
            self.manager.remove_constraint(self.constraint.tag)
            
            args = [new_master, new_slaves]
            if isinstance(self.constraint, equalDOF):
                new_dofs = list(map(int, self.dofs_input.text().split()))
                args.append(new_dofs)
            elif isinstance(self.constraint, rigidLink):
                args.insert(0, self.type_combo.currentText())
            elif isinstance(self.constraint, rigidDiaphragm):
                args.insert(0, int(self.direction_input.text()))
            
            self.manager.create_constraint(type(self.constraint).__name__, *args)
            self.accept()
            
        except ValueError as e:
            QMessageBox.warning(self, "Input Error", f"Invalid input: {str(e)}")

if __name__ == '__main__':
    from qtpy.QtWidgets import QApplication
    import sys
    
    # Create some sample constraints for testing
    mp_manager = mpConstraintManager()
    
    # Create equalDOF constraints
    mp_manager.create_equal_dof(master_node=1, slave_nodes=[2, 3], dofs=[1, 2, 3])
    mp_manager.create_equal_dof(master_node=5, slave_nodes=[6], dofs=[4, 5, 6])
    
    # Create rigidLink constraints
    mp_manager.create_rigid_link(type_str="bar", master_node=10, slave_nodes=[11, 12, 13])
    mp_manager.create_rigid_link(type_str="beam", master_node=20, slave_nodes=[21, 22])
    
    # Create rigidDiaphragm constraints
    mp_manager.create_rigid_diaphragm(direction=3, master_node=30, slave_nodes=[31, 32, 33, 34])
    mp_manager.create_rigid_diaphragm(direction=2, master_node=40, slave_nodes=[41, 42, 43])
    
    # Start the application
    app = QApplication(sys.argv)
    manager_tab = MPConstraintManagerTab()
    manager_tab.show()
    sys.exit(app.exec())


from qtpy.QtWidgets import QComboBox, QDoubleSpinBox
from .base_creator import DRMCreatorDialog


class SurfaceWaveCreator(DRMCreatorDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Create Surface Wave")
        self.setup_form()
        
    def setup_form(self):
        # Add your surface wave specific parameters here
        self.wave_type = QComboBox()
        self.wave_type.addItems(['Rayleigh', 'Love'])
        self.form_layout.addRow("Wave Type:", self.wave_type)
        
        self.frequency = QDoubleSpinBox()
        self.frequency.setRange(0.1, 100.0)
        self.frequency.setValue(1.0)
        self.form_layout.addRow("Frequency (Hz):", self.frequency)
        
        self.wavelength = QDoubleSpinBox()
        self.wavelength.setRange(1.0, 1000.0)
        self.wavelength.setValue(100.0)
        self.form_layout.addRow("Wavelength (m):", self.wavelength)
    
    def get_parameters(self):
        return {
            'wave_type': self.wave_type.currentText(),
            'frequency': self.frequency.value(),
            'wavelength': self.wavelength.value()
        }
    
    def create_load(self):
        params = self.get_parameters()
        # Implement your surface wave creation logic here
        print(f"Creating surface wave with parameters: {params}")
        # Return the created wave or None if failed
        return params